# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import numbers
from typing import Optional, Sequence

import numpy as np
import onnx
import onnx_ir as ir

from onnxscript import tensor


def external_tensor(
    name: str,
    data_type: int,
    dims: Sequence[int],
    location: str,
    offset: Optional[int] = None,
    length: Optional[int] = None,
    checksum: Optional[str] = None,
    basepath: Optional[str] = None,
) -> onnx.TensorProto:
    """Create a TensorProto referencing externally stored tensor-data.

    Args:
        name: name of the tensor
        data_type: data type of tensor element
        dims: shape of the tensor
        location: location of the external file (relative path)
        offset: offset in the file where the tensor-data starts
        length: number of bytes containing the data
        checksum: SHA1 digest of the file
        basepath: basepath combined with location to form the full path

    Returns:
        TensorProto

    See https://github.com/onnx/onnx/blob/main/docs/ExternalData.md for more details.
    """
    tensor_proto = onnx.TensorProto()
    tensor_proto.name = name
    tensor_proto.data_type = data_type
    tensor_proto.dims.extend(dims)
    tensor_proto.data_location = onnx.TensorProto.EXTERNAL

    def add(k, v):
        entry = tensor_proto.external_data.add()
        entry.key = k
        entry.value = str(v)

    add("location", location)
    if offset is not None:
        add("offset", int(offset))
    if length is not None:
        add("length", int(length))
    if checksum is not None:
        add("checksum", checksum)
    if basepath is not None:
        add("basepath", basepath)
    return tensor_proto


def value_to_type_proto(val):
    """Return the ONNX type of a python-value."""
    if isinstance(val, (np.ndarray, tensor.Tensor)):
        elem_type = onnx.helper.np_dtype_to_tensor_dtype(val.dtype)  # noqa: TID251
        shape = val.shape
        return onnx.helper.make_tensor_type_proto(elem_type, shape)  # noqa: TID251
    if isinstance(val, int):
        return onnx.helper.make_tensor_type_proto(onnx.TensorProto.INT32, [])  # noqa: TID251
    if isinstance(val, (float, np.float32)):
        return onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, [])  # noqa: TID251
    if isinstance(val, list):
        if len(val) > 0:
            return onnx.helper.make_sequence_type_proto(value_to_type_proto(val[0]))  # noqa: TID251
        # Edge-case. Cannot determine a suitable ONNX type for an empty list.
        # Should be using a typed-value instead.
        # Treated as a sequence of tensors of float-type.
        return onnx.helper.make_sequence_type_proto(  # noqa: TID251
            onnx.helper.make_tensor_type_proto(onnx.TensorProto.FLOAT, None)  # noqa: TID251
        )
    if isinstance(val, numbers.Number):
        nparray = np.array(val)
        elem_type = onnx.helper.np_dtype_to_tensor_dtype(nparray.dtype)  # noqa: TID251
        return onnx.helper.make_tensor_type_proto(elem_type, [])  # noqa: TID251
    raise ValueError(f"Value of type {type(val)} is invalid as an ONNX input/output.")


def value_to_type(val):
    """Return an ir.Value representation of a python-value."""
    if isinstance(val, (np.ndarray, tensor.Tensor)):
        elem_type = onnx.helper.np_dtype_to_tensor_dtype(val.dtype)  # noqa: TID251
        shape = val.shape
        return (ir.TensorType(elem_type), shape)
    elif isinstance(val, int):
        elem_type = onnx.TensorProto.INT32
        shape = []
        return (ir.TensorType(elem_type), shape)
    elif isinstance(val, (float, np.float32)):
        elem_type = onnx.TensorProto.FLOAT
        shape = []
        return (ir.TensorType(elem_type), shape)
    elif isinstance(val, list):
        if len(val) > 0:
            type, shape = value_to_type(val[0])
            return ir.SequenceType(type), shape
        # Edge-case. Cannot determine a suitable ONNX type for an empty list.
        # Should be using a typed-value instead.
        # Treated as a sequence of tensors of float-type.
        return ir.SequenceType(ir.TensorType(onnx.TensorProto.FLOAT)), None
    if isinstance(val, numbers.Number):
        nparray = np.array(val)
        elem_type = onnx.helper.np_dtype_to_tensor_dtype(nparray.dtype)  # noqa: TID251
        return ir.TensorType(elem_type), []
    raise ValueError(f"Value of type {type(val)} is invalid as an ONNX input/output.")


def value_to_ir_value(name: str, val) -> ir.Value:
    """Return an ir.Value representation of a python-value."""
    type, shape = value_to_type(val)
    return ir.Value(name=name, type=type, shape=shape)


def values_to_value_infos(name_values):
    """Create a list of ValueInfoProto from a list of (name, value) pairs,
    skipping any None values.
    """
    return [
        onnx.helper.make_value_info(name, value_to_type_proto(val))  # noqa: TID251
        for (name, val) in name_values
        if val is not None
    ]


def values_to_ir_values(name_values):
    """Create a list of ir.Value from a list of (name, value) pairs,
    skipping any None values.
    """
    return [value_to_ir_value(name, val) for (name, val) in name_values if val is not None]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# mypy: disable-error-code="misc,arg-type,type-arg,valid-type,assignment,return-value"
# pylint: disable=unused-argument
"""quantized_decomposed ops defined in https://github.com/pytorch/pytorch/blob/main/torch/ao/quantization/fx/_decomposed.py

- No inplace operators.
- All functions should not have the script() decorator. This is because
    we want to delay the compilation of the function.
"""

from __future__ import annotations

from onnxscript.function_libs.torch_lib.ops import common
from onnxscript.function_libs.torch_lib.registration import torch_op
from onnxscript.onnx_opset import opset18 as op
from onnxscript.onnx_types import TensorType


@torch_op(
    (
        "quantized_decomposed::quantize_per_tensor",
        "quantized_decomposed::quantize_per_tensor.tensor",
        "quantized_decomposed::quantize_per_tensor.tensor2",
    ),
    trace_only=True,
)
def quantized_decomposed_quantize_per_tensor(
    input: TensorType,
    scale: float,
    zero_point: int,
    quant_min: int,
    quant_max: int,
    dtype: int,
) -> TensorType:
    # TODO(justinchuby): Use dtype when we use opset 21
    return op.QuantizeLinear(input, scale, common.constant(zero_point, dtype=dtype))


@torch_op(
    (
        "quantized_decomposed::dequantize_per_tensor",
        "quantized_decomposed::dequantize_per_tensor.tensor",
        "quantized_decomposed::dequantize_per_tensor.tensor2",
    ),
    trace_only=True,
)
def quantized_decomposed_dequantize_per_tensor(
    input: TensorType,
    scale: float,
    zero_point: int,
    quant_min: int,
    quant_max: int,
    dtype: int,
    out_dtype: int = -1,
) -> TensorType:
    # TODO(justinchuby): Use dtype when we use opset 21
    dequantized = op.DequantizeLinear(input, scale, common.constant(zero_point, dtype=dtype))
    if out_dtype in (-1, None):
        # out_dtype can be None as well
        return dequantized
    assert out_dtype > 0, f"out_dtype must be -1 or > 0 not {out_dtype}"
    return op.Cast(dequantized, to=out_dtype)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import logging
from typing import Callable

import onnx_ir as ir
import onnx_ir.passes.common as common_passes

from onnxscript import rewriter
from onnxscript.optimizer import _constant_folding

logger = logging.getLogger(__name__)


def optimize_ir(
    model: ir.Model,
    num_iterations: int = 2,
    *,
    onnx_shape_inference: bool = True,
    stop_if_no_change: bool = True,
    input_size_limit: int = _constant_folding.DEFAULT_CONSTANT_FOLD_INPUT_SIZE_LIMIT,
    output_size_limit: int = _constant_folding.DEFAULT_CONSTANT_FOLD_OUTPUT_SIZE_LIMIT,
    should_fold: Callable[[ir.Node], bool | None] = lambda node: None,
    inline: bool = True,
) -> None:
    """Optimizes a model.

    Args:
        model: The model to be optimized.
        num_iterations: Number of times the optimization loop is repeated.
        onnx_shape_inference: Applies node-level shape-inference as part of optimization
        stop_if_no_change: Stop the optimization loop if no change is detected in an iteration.
        input_size_limit: Will not apply constant folding to ops with any input of size
            greater than this. Does not apply to special ops like Shape() and Size().
        output_size_limit: Will not rewrite any foldable-op into a Constant op if the size
            of the output tensor is greater than this.
        should_fold: An optional function that takes a node and returns True if
            the node should be considered for folding.
            The function should return True/False value to indicate if this particular
            node should be folded, or None to use the default folding rules.
        inline: If True, inlines all functions in the model.
    """
    passes = [
        ir.passes.PassManager(
            [
                _constant_folding.FoldConstantsPass(
                    shape_inference=onnx_shape_inference,
                    input_size_limit=input_size_limit,
                    output_size_limit=output_size_limit,
                    should_fold=should_fold,
                ),
                rewriter.RewritePass(rewriter._DEFAULT_REWRITE_RULES),
                common_passes.RemoveUnusedNodesPass(),
                common_passes.RemoveUnusedFunctionsPass(),
                common_passes.RemoveUnusedOpsetsPass(),
            ],
            steps=num_iterations,
            early_stop=stop_if_no_change,
        ),
        common_passes.RemoveUnusedNodesPass(),
        common_passes.LiftConstantsToInitializersPass(lift_all_constants=True, size_limit=0),
        common_passes.LiftSubgraphInitializersToMainGraphPass(),
        common_passes.DeduplicateInitializersPass(),
        common_passes.CommonSubexpressionEliminationPass(),
        common_passes.OutputFixPass(),
        common_passes.NameFixPass(),
    ]
    if inline:
        # Inline all functions first before optimizing
        passes = [common_passes.InlinePass(), *passes]
    optimizer_pass = ir.passes.Sequential(*passes)
    assert optimizer_pass.in_place
    result = optimizer_pass(model)
    assert result.model is model



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest
from typing import Any, Optional, Sequence, TypeVar, Union

import parameterized

import onnxscript
from onnxscript import FLOAT, INT64, ir
from onnxscript.ir import _schemas

_TestTypeVarConstraints = TypeVar("_TestTypeVarConstraints", INT64, FLOAT)
_TestTypeVarOneBound = TypeVar("_TestTypeVarOneBound", bound=INT64)
_TestTypeVarTwoBound = TypeVar("_TestTypeVarTwoBound", bound=Union[INT64, FLOAT])


class TypeConversionFunctionsTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (
                "tensor_type_all",
                onnxscript.onnx_types.TensorType,
                {ir.TensorType(dtype) for dtype in ir.DataType},
            ),
            ("tensor_type", INT64, {ir.TensorType(ir.DataType.INT64)}),
            (
                "tensor_type_union",
                Union[INT64, FLOAT],
                {ir.TensorType(ir.DataType.INT64), ir.TensorType(ir.DataType.FLOAT)},
            ),
            (
                "tensor_type_variadic_shape",
                INT64[...],
                {ir.TensorType(ir.DataType.INT64)},
            ),
            ("tensor_type_shape", INT64[10], {ir.TensorType(ir.DataType.INT64)}),
            (
                "type_var_constraints",
                _TestTypeVarConstraints,
                {ir.TensorType(ir.DataType.INT64), ir.TensorType(ir.DataType.FLOAT)},
            ),
            (
                "type_bound_one",
                _TestTypeVarOneBound,
                {ir.TensorType(ir.DataType.INT64)},
            ),
            (
                "type_bound_two",
                _TestTypeVarTwoBound,
                {ir.TensorType(ir.DataType.INT64), ir.TensorType(ir.DataType.FLOAT)},
            ),
            (
                "optional_tensor_type_all",
                Optional[onnxscript.onnx_types.TensorType],
                {ir.TensorType(dtype) for dtype in ir.DataType},
            ),
            (
                "optional_tensor_type",
                Optional[INT64],
                {ir.TensorType(ir.DataType.INT64)},
            ),
            (
                "optional_tensor_type_union",
                Optional[Union[INT64, FLOAT]],
                {ir.TensorType(ir.DataType.INT64), ir.TensorType(ir.DataType.FLOAT)},
            ),
            (
                "optional_tensor_type_variadic_shape",
                Optional[INT64[...]],
                {ir.TensorType(ir.DataType.INT64)},
            ),
            (
                "optional_tensor_type_shape",
                Optional[INT64[10]],
                {ir.TensorType(ir.DataType.INT64)},
            ),
            (
                "optional_type_var_constraints",
                Optional[_TestTypeVarConstraints],
                {ir.TensorType(ir.DataType.INT64), ir.TensorType(ir.DataType.FLOAT)},
            ),
            (
                "optional_type_bound_one",
                Optional[_TestTypeVarOneBound],
                {ir.TensorType(ir.DataType.INT64)},
            ),
            (
                "optional_type_bound_two",
                Optional[_TestTypeVarTwoBound],
                {ir.TensorType(ir.DataType.INT64), ir.TensorType(ir.DataType.FLOAT)},
            ),
            (
                "sequence_type_all",
                Sequence[onnxscript.onnx_types.TensorType],
                {ir.SequenceType(ir.TensorType(dtype)) for dtype in ir.DataType},
            ),
            (
                "sequence_type",
                Sequence[INT64],
                {ir.SequenceType(ir.TensorType(ir.DataType.INT64))},
            ),
            (
                "union_sequence_type",
                Union[Sequence[INT64], Sequence[FLOAT]],
                {
                    ir.SequenceType(ir.TensorType(ir.DataType.INT64)),
                    ir.SequenceType(ir.TensorType(ir.DataType.FLOAT)),
                },
            ),
            (
                "sequence_type_variadic_shape",
                Sequence[INT64[...]],
                {ir.SequenceType(ir.TensorType(ir.DataType.INT64))},
            ),
            (
                "sequence_type_shape",
                Sequence[INT64[10]],
                {ir.SequenceType(ir.TensorType(ir.DataType.INT64))},
            ),
            (
                "sequence_type_var_constraints",
                Sequence[_TestTypeVarConstraints],
                {
                    ir.SequenceType(ir.TensorType(ir.DataType.INT64)),
                    ir.SequenceType(ir.TensorType(ir.DataType.FLOAT)),
                },
            ),
            (
                "sequence_type_bound_one",
                Sequence[_TestTypeVarOneBound],
                {ir.SequenceType(ir.TensorType(ir.DataType.INT64))},
            ),
            (
                "sequence_type_bound_two",
                Sequence[_TestTypeVarTwoBound],
                {
                    ir.SequenceType(ir.TensorType(ir.DataType.INT64)),
                    ir.SequenceType(ir.TensorType(ir.DataType.FLOAT)),
                },
            ),
        ]
    )
    def test_pytype_to_ir_type(self, _, pytype: Any, expected: set[ir.TypeProtocol]):
        self.assertEqual(_schemas._get_allowed_types_from_type_annotation(pytype), expected)  # pylint: disable=protected-access

    @parameterized.parameterized.expand(
        [
            ("type_var", _TestTypeVarConstraints, "_TestTypeVarConstraints"),
            ("type_var_bound", _TestTypeVarOneBound, "_TestTypeVarOneBound"),
            (
                "optional_type_var",
                Optional[_TestTypeVarOneBound],
                "_TestTypeVarOneBound",
            ),
            (
                "sequence_type_var",
                Sequence[_TestTypeVarOneBound],
                "Sequence__TestTypeVarOneBound",
            ),
            ("normal_type", INT64, None),
            ("union_type", Union[INT64, FLOAT], None),
            ("optional_type", Optional[INT64], None),
            ("sequence_type", Sequence[INT64], None),
            ("optional_sequence_type", Optional[Sequence[INT64]], None),
            ("optional_union_type", Optional[Union[INT64, FLOAT]], None),
        ]
    )
    def test_get_type_constraint_name(self, _: str, pytype: Any, expected: str | None):
        self.assertEqual(_schemas._get_type_constraint_name(pytype), expected)  # pylint: disable=protected-access


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from __future__ import annotations

import unittest

import onnx_ir as ir

from onnxscript.onnx_types import FLOAT, INT64


class TensorTypeToIrTest(unittest.TestCase):
    """Tests for TensorType.to_ir_type_and_shape()."""

    def test_scalar_type(self):
        """FLOAT (no subscript) maps to rank-0 tensor (empty shape)."""
        ts = FLOAT.to_ir_type_and_shape()
        self.assertIsInstance(ts, ir.TypeAndShape)
        self.assertEqual(ts.type, ir.TensorType(ir.DataType.FLOAT))
        self.assertIsNotNone(ts.shape)
        self.assertEqual(len(ts.shape), 0)

    def test_unknown_rank(self):
        """FLOAT[...] maps to unknown-rank (shape=None)."""
        ts = FLOAT[...].to_ir_type_and_shape()
        self.assertIsInstance(ts, ir.TypeAndShape)
        self.assertIsNone(ts.shape)

    def test_single_dim(self):
        """FLOAT[1024] maps to a 1-D tensor with dimension 1024."""
        ts = FLOAT[1024].to_ir_type_and_shape()
        self.assertIsNotNone(ts.shape)
        self.assertEqual(len(ts.shape), 1)
        self.assertEqual(ts.shape[0], 1024)

    def test_multi_dim_int(self):
        """FLOAT[3, 4] maps to a 2-D tensor with dims (3, 4)."""
        ts = FLOAT[3, 4].to_ir_type_and_shape()
        self.assertIsNotNone(ts.shape)
        self.assertEqual(len(ts.shape), 2)
        self.assertEqual(ts.shape[0], 3)
        self.assertEqual(ts.shape[1], 4)

    def test_symbolic_dims(self):
        """FLOAT['M', 'N'] maps to a 2-D tensor with symbolic dims."""
        ts = FLOAT["M", "N"].to_ir_type_and_shape()
        self.assertIsNotNone(ts.shape)
        self.assertEqual(len(ts.shape), 2)

    def test_other_dtype(self):
        """INT64[...] preserves the correct dtype."""
        ts = INT64[...].to_ir_type_and_shape()
        self.assertEqual(ts.type, ir.TensorType(ir.DataType.INT64))


if __name__ == "__main__":
    unittest.main()



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# mypy: disable-error-code="misc,arg-type,type-arg,valid-type,assignment,return-value"
"""torch.ops.aten operators under the `prims` module.

- No inplace operators.
- All functions should not have the script() decorator. This is because
    we want to delay the compilation of the function.
"""

from __future__ import annotations

from typing import Optional, Sequence

from onnxscript import INT64
from onnxscript.function_libs.torch_lib.ops import common as common_ops
from onnxscript.function_libs.torch_lib.registration import torch_op
from onnxscript.function_libs.torch_lib.tensor_typing import RealType, TTensor
from onnxscript.onnx_opset import opset18 as op
from onnxscript.onnx_types import BOOL, TensorType


@torch_op("prims::abs", trace_only=True)
def prims_abs(self: TTensor) -> TTensor:
    """abs(Tensor self) -> Tensor"""

    return op.Abs(self)


@torch_op("prims::acos", trace_only=True)
def prims_acos(self: TensorType) -> TensorType:
    """acos(Tensor self) -> Tensor"""

    return op.Acos(self)


@torch_op("prims::acosh", trace_only=True)
def prims_acosh(self: TensorType) -> TensorType:
    """acosh(Tensor self) -> Tensor"""

    return op.Acosh(self)


@torch_op("prims::add", trace_only=True)
def prims_add(self: TTensor, other: TTensor) -> TTensor:
    """add(Tensor self, Tensor other) -> Tensor"""

    return op.Add(self, other)


def prims_amax(
    inp: TensorType, dims: Optional[Sequence[int]], output_dtype: Optional[int] = None
) -> TensorType:
    """amax(Tensor inp, int[]? dims, *, ScalarType? output_dtype=None) -> Tensor"""

    raise NotImplementedError()


def prims_amin(
    inp: TensorType, dims: Optional[Sequence[int]], output_dtype: Optional[int] = None
) -> TensorType:
    """amin(Tensor inp, int[]? dims, *, ScalarType? output_dtype=None) -> Tensor"""

    raise NotImplementedError()


def prims_as_strided(
    a: TensorType, size: INT64, stride: INT64, storage_offset: INT64
) -> TensorType:
    """as_strided(Tensor a, SymInt[] size, SymInt[] stride, SymInt storage_offset) -> Tensor"""

    raise NotImplementedError()


def prims_as_strided_scatter(
    self: TensorType, src: TensorType, size: INT64, stride: INT64, storage_offset: INT64
) -> TensorType:
    """as_strided_scatter(Tensor self, Tensor src, SymInt[] size, SymInt[] stride, SymInt storage_offset) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::asin", trace_only=True)
def prims_asin(self: TTensor) -> TTensor:
    """asin(Tensor self) -> Tensor"""

    return op.Asin(self)


@torch_op("prims::asinh", trace_only=True)
def prims_asinh(self: TTensor) -> TTensor:
    """asinh(Tensor self) -> Tensor"""

    return op.Asinh(self)


@torch_op("prims::atan", trace_only=True)
def prims_atan(self: TTensor) -> TTensor:
    """atan(Tensor self) -> Tensor"""

    return op.Atan(self)


def prims_atan2(self: TensorType, other: TensorType) -> TensorType:
    """atan2(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::atanh", trace_only=True)
def prims_atanh(self: TTensor) -> TTensor:
    """atanh(Tensor self) -> Tensor"""

    return op.Atanh(self)


def prims_bessel_i0(self: TensorType) -> TensorType:
    """bessel_i0(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_bessel_i0e(self: TensorType) -> TensorType:
    """bessel_i0e(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_bessel_i1(self: TensorType) -> TensorType:
    """bessel_i1(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_bessel_i1e(self: TensorType) -> TensorType:
    """bessel_i1e(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_bessel_j0(self: TensorType) -> TensorType:
    """bessel_j0(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_bessel_j1(self: TensorType) -> TensorType:
    """bessel_j1(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_bitwise_and(self: TensorType, other: TensorType) -> TensorType:
    """bitwise_and(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_bitwise_not(self: TensorType) -> TensorType:
    """bitwise_not(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_bitwise_or(self: TensorType, other: TensorType) -> TensorType:
    """bitwise_or(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_bitwise_xor(self: TensorType, other: TensorType) -> TensorType:
    """bitwise_xor(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::broadcast_in_dim", trace_only=True)
def prims_broadcast_in_dim(
    a: TensorType, shape: Sequence[INT64], broadcast_dimensions: Sequence[int]
) -> TensorType:
    """broadcast_in_dim(Tensor(a) a, SymInt[] shape, int[] broadcast_dimensions) -> Tensor(a)"""

    target_rank = len(shape)

    if not broadcast_dimensions:
        # Special case: no broadcast dimensions - all target dims should be 1
        return op.Expand(a, common_ops.merge_dims(shape))

    # Create base shape of all 1s
    ones = [1] * target_rank

    # For each broadcast dimension, we'll replace the 1 with the actual input dimension
    # Since broadcast_dimensions is compile-time known, we can do this with individual operations
    intermediate_shape = ones

    for i, broadcast_dim in enumerate(broadcast_dimensions):
        # Get the input dimension value
        input_dim_value = op.Shape(a, start=i, end=i + 1)
        intermediate_shape[broadcast_dim] = input_dim_value

    # Reshape input to intermediate shape and expand to target
    reshaped = op.Reshape(a, common_ops.merge_dims(intermediate_shape))
    return op.Expand(reshaped, shape)


def prims_cat(tensors: Sequence[TensorType], dim: int) -> TensorType:
    """cat(Tensor[] tensors, int dim) -> Tensor"""

    raise NotImplementedError()


def prims_cbrt(self: TensorType) -> TensorType:
    """cbrt(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::ceil", trace_only=True)
def prims_ceil(self: TTensor) -> TTensor:
    """ceil(Tensor self) -> Tensor"""

    return op.Ceil(self)


def prims_clone(self: TensorType, memory_format: Optional[str] = None) -> TensorType:
    """clone(Tensor self, *, MemoryFormat? memory_format=None) -> Tensor"""

    raise NotImplementedError()


def prims_collapse_view(a: TensorType, start: int, end: int) -> TensorType:
    """collapse_view(Tensor(a) a, int start, int end) -> Tensor(a)"""

    raise NotImplementedError()


def prims_conj(a: TensorType) -> TensorType:
    """conj(Tensor(a) a) -> Tensor(a)"""

    raise NotImplementedError()


def prims_conj_physical(self: TensorType) -> TensorType:
    """conj_physical(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::convert_element_type", trace_only=True)
def prims_convert_element_type(a: RealType, dtype: int) -> RealType:
    """convert_element_type(Tensor a, ScalarType dtype) -> Tensor"""

    # Set trace_only=True because different if branches return different dtypes
    # which is not supported in an ONNX function
    return common_ops.cast_to(a, dtype)


def prims_copy_strided(a: TensorType, stride: INT64) -> TensorType:
    """copy_strided(Tensor a, SymInt[] stride) -> Tensor"""

    raise NotImplementedError()


def prims_copy_to(a: TensorType, b: TensorType) -> TensorType:
    """copy_to(Tensor a, Tensor b) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::cos", trace_only=True)
def prims_cos(self: TTensor) -> TTensor:
    """cos(Tensor self) -> Tensor"""

    return op.Cos(self)


@torch_op("prims::cosh", trace_only=True)
def prims_cosh(self: TTensor) -> TTensor:
    """cosh(Tensor self) -> Tensor"""

    return op.Cosh(self)


@torch_op("prims::device_put")
def prims_device_put(
    a: TTensor,
    device: str = "unspecified",  # pylint: disable=unused-argument
) -> TTensor:
    """device_put(Tensor a, Device device) -> Tensor"""

    # ONNX does not have the notion of a "device". So we just return the input
    return op.Identity(a)


def prims_digamma(self: TensorType) -> TensorType:
    """digamma(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::div", trace_only=True)
def prims_div(self: TTensor, other: TTensor) -> TTensor:
    """div(Tensor self, Tensor other) -> Tensor"""

    return op.Div(self, other)


def prims_empty(shape: INT64, dtype: int, device: str, requires_grad: bool) -> TensorType:
    """empty(SymInt[] shape, *, ScalarType dtype, Device device, bool requires_grad) -> Tensor"""

    raise NotImplementedError()


def prims_empty_strided(
    shape: INT64, strides: INT64, dtype: int, device: str, requires_grad: bool
) -> TensorType:
    """empty_strided(SymInt[] shape, SymInt[] strides, *, ScalarType dtype, Device device, bool requires_grad) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::eq", trace_only=True)
def prims_eq(self: TTensor, other: TTensor) -> TTensor:
    """eq(Tensor self, Tensor other) -> Tensor"""

    return op.Equal(self, other)


@torch_op("prims::erf", trace_only=True)
def prims_erf(self: TTensor) -> TTensor:
    """erf(Tensor self) -> Tensor"""

    return op.Erf(self)


def prims_erf_inv(self: TensorType) -> TensorType:
    """erf_inv(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_erfc(self: TensorType) -> TensorType:
    """erfc(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_erfcx(self: TensorType) -> TensorType:
    """erfcx(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::exp", trace_only=True)
def prims_exp(self: TTensor) -> TTensor:
    """exp(Tensor self) -> Tensor"""

    return op.Exp(self)


def prims_exp2(self: TensorType) -> TensorType:
    """exp2(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_expm1(self: TensorType) -> TensorType:
    """expm1(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_fft_c2c(self: TensorType, dim: Sequence[int], forward: bool) -> TensorType:
    """fft_c2c(Tensor self, *, int[] dim, bool forward) -> Tensor"""

    raise NotImplementedError()


def prims_fft_c2r(self: TensorType, dim: Sequence[int], last_dim_size: INT64) -> TensorType:
    """fft_c2r(Tensor self, *, int[] dim, SymInt last_dim_size) -> Tensor"""

    raise NotImplementedError()


def prims_fft_r2c(self: TensorType, dim: Sequence[int], onesided: bool) -> TensorType:
    """fft_r2c(Tensor self, *, int[] dim, bool onesided) -> Tensor"""

    raise NotImplementedError()


def prims_fill(self: TensorType, value: float) -> TensorType:
    """fill(Tensor self, Scalar value) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::floor", trace_only=True)
def prims_floor(self: TTensor) -> TTensor:
    """floor(Tensor self) -> Tensor"""

    return op.Floor(self)


def prims_fmax(self: TensorType, other: TensorType) -> TensorType:
    """fmax(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_fmin(self: TensorType, other: TensorType) -> TensorType:
    """fmin(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_fmod(self: TensorType, other: TensorType) -> TensorType:
    """fmod(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_full(
    shape: INT64, fill_value: float, dtype: int, device: str, requires_grad: bool
) -> TensorType:
    """full(SymInt[] shape, Scalar fill_value, *, ScalarType dtype, Device device, bool requires_grad) -> Tensor"""

    raise NotImplementedError()


def prims_full_like(
    a: TensorType, fill_value: float, dtype: int, device: str, requires_grad: bool
) -> TensorType:
    """full_like(Tensor a, Scalar fill_value, *, ScalarType dtype, Device device, bool requires_grad) -> Tensor"""

    raise NotImplementedError()


def prims_gcd(self: TensorType, other: TensorType) -> TensorType:
    """gcd(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::ge", trace_only=True)
def prims_ge(self: TTensor, other: TTensor) -> TTensor:
    """ge(Tensor self, Tensor other) -> Tensor"""

    return op.GreaterOrEqual(self, other)


@torch_op("prims::gt", trace_only=True)
def prims_gt(self: TTensor, other: TTensor) -> TTensor:
    """gt(Tensor self, Tensor other) -> Tensor"""

    return op.Greater(self, other)


def prims_hypot(self: TensorType, other: TensorType) -> TensorType:
    """hypot(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_igamma(self: TensorType, other: TensorType) -> TensorType:
    """igamma(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_igammac(self: TensorType, other: TensorType) -> TensorType:
    """igammac(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_imag(self: TensorType) -> TensorType:
    """imag(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_iota(
    length: INT64, start: INT64, step: INT64, dtype: int, device: str, requires_grad: bool
) -> TensorType:
    """iota(SymInt length, *, SymInt start, SymInt step, ScalarType dtype, Device device, bool requires_grad) -> Tensor"""

    raise NotImplementedError()


def prims_isfinite(self: TensorType) -> TensorType:
    """isfinite(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_item(a: TensorType) -> float:
    """item(Tensor a) -> Scalar"""

    raise NotImplementedError()


@torch_op("prims::le", trace_only=True)
def prims_le(self: TensorType, other: TensorType) -> TensorType:
    """le(Tensor self, Tensor other) -> Tensor"""

    return op.LessOrEqual(self, other)


def prims_lgamma(self: TensorType) -> TensorType:
    """lgamma(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::log", trace_only=True)
def prims_log(self: TensorType) -> TensorType:
    """log(Tensor self) -> Tensor"""

    return op.Log(self)


def prims_log10(self: TensorType) -> TensorType:
    """log10(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_log1p(self: TensorType) -> TensorType:
    """log1p(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_log2(self: TensorType) -> TensorType:
    """log2(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::lt", trace_only=True)
def prims_lt(self: TensorType, other: TensorType) -> TensorType:
    """lt(Tensor self, Tensor other) -> Tensor"""

    return op.Less(self, other)


def prims_maximum(self: TensorType, other: TensorType) -> TensorType:
    """maximum(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_maximum_value(dtype: int) -> float:
    """maximum_value(ScalarType dtype) -> Scalar"""

    raise NotImplementedError()


def prims_minimum(self: TensorType, other: TensorType) -> TensorType:
    """minimum(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_minium_value(dtype: int) -> float:
    """minium_value(ScalarType dtype) -> Scalar"""

    raise NotImplementedError()


@torch_op("prims::mul", trace_only=True)
def prims_mul(self: TTensor, other: TTensor) -> TTensor:
    """mul(Tensor self, Tensor other) -> Tensor"""

    return op.Mul(self, other)


def prims_ndtri(self: TensorType) -> TensorType:
    """ndtri(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::ne", trace_only=True)
def prims_ne(self: TTensor, other: TTensor) -> TTensor:
    """ne(Tensor self, Tensor other) -> Tensor"""

    return op.Not(op.Equal(self, other))


@torch_op("prims::neg", trace_only=True)
def prims_neg(self: TTensor) -> TTensor:
    """neg(Tensor self) -> Tensor"""

    return op.Neg(self)


def prims_nextafter(self: TensorType, other: TensorType) -> TensorType:
    """nextafter(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_normal(
    shape: INT64, mean: float, std: float, dtype: int, device: str, requires_grad: bool
) -> TensorType:
    """normal(SymInt[] shape, *, Scalar mean, Scalar std, ScalarType dtype, Device device, bool requires_grad) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::pow", trace_only=True)
def prims_pow(self: TTensor, other: TTensor) -> TTensor:
    """pow(Tensor self, Tensor other) -> Tensor"""

    return op.Pow(self, other)


def prims_prod(
    inp: TensorType, dims: Optional[Sequence[int]], output_dtype: Optional[int] = None
) -> TensorType:
    """prod(Tensor inp, int[]? dims, *, ScalarType? output_dtype=None) -> Tensor"""

    raise NotImplementedError()


def prims_real(self: TensorType) -> TensorType:
    """real(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_reciprocal(self: TensorType) -> TensorType:
    """reciprocal(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_remainder(self: TensorType, other: TensorType) -> TensorType:
    """remainder(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::reshape", trace_only=True)
def prims_reshape(a: TTensor, shape: INT64) -> TTensor:
    """reshape(Tensor a, SymInt[] shape) -> Tensor"""

    return op.Reshape(a, shape)


@torch_op("prims::resize", trace_only=True)
def prims_resize(a: TensorType, shape: INT64) -> TensorType:
    """resize(Tensor a, SymInt[] shape) -> Tensor"""

    return op.Expand(a, shape)


def prims_rev(a: TensorType, dims: Sequence[int]) -> TensorType:
    """rev(Tensor a, int[] dims) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::round", trace_only=True)
def prims_round(self: TensorType) -> TensorType:
    """round(Tensor self) -> Tensor"""

    return op.Round(self)


def prims_rsqrt(self: TensorType) -> TensorType:
    """rsqrt(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_scalar_tensor(
    s: float, dtype: Optional[int] = None, device: Optional[str] = None
) -> TensorType:
    """scalar_tensor(Scalar s, *, ScalarType? dtype=None, Device? device=None) -> Tensor"""

    raise NotImplementedError()


def prims_shift_left(self: TensorType, other: TensorType) -> TensorType:
    """shift_left(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_shift_right_arithmetic(self: TensorType, other: TensorType) -> TensorType:
    """shift_right_arithmetic(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()


def prims_sign(self: TensorType) -> TensorType:
    """sign(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_signbit(self: TensorType) -> TensorType:
    """signbit(Tensor self) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::sin", trace_only=True)
def prims_sin(self: TTensor) -> TTensor:
    """sin(Tensor self) -> Tensor"""

    return op.Sin(self)


@torch_op("prims::sinh", trace_only=True)
def prims_sinh(self: TTensor) -> TTensor:
    """sinh(Tensor self) -> Tensor"""

    return op.Sinh(self)


def prims_slice(
    a: TensorType, start_indices: INT64, limit_indices: INT64, strides: Optional[INT64] = None
) -> TensorType:
    """slice(Tensor(a) a, SymInt[] start_indices, SymInt[] limit_indices, SymInt[]? strides=None) -> Tensor(a)"""

    raise NotImplementedError()


def prims_slice_in_dim(
    a: TensorType, start_index: INT64, limit_index: INT64, stride: int = 1, axis: int = 0
) -> TensorType:
    """slice_in_dim(Tensor(a) a, SymInt start_index, SymInt limit_index, int stride=1, int axis=0) -> Tensor(a)"""

    raise NotImplementedError()


def prims_spherical_bessel_j0(self: TensorType) -> TensorType:
    """spherical_bessel_j0(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_split_dim(a: TensorType, dim: int, outer_length: INT64) -> TensorType:
    """split_dim(Tensor(a) a, int dim, SymInt outer_length) -> Tensor(a)"""

    raise NotImplementedError()


@torch_op("prims::sqrt", trace_only=True)
def prims_sqrt(self: TTensor) -> TTensor:
    """sqrt(Tensor self) -> Tensor"""

    return op.Sqrt(self)


@torch_op("prims::squeeze", trace_only=True)
def prims_squeeze(a: TTensor, dimensions: Sequence[int]) -> TTensor:
    """squeeze(Tensor(a) a, int[] dimensions) -> Tensor(a)"""

    return op.Squeeze(a, axes=dimensions)


@torch_op("prims::sub", trace_only=True)
def prims_sub(self: TTensor, other: TTensor) -> TTensor:
    """sub(Tensor self, Tensor other) -> Tensor"""

    return op.Sub(self, other)


@torch_op("prims::sum", trace_only=True)
def prims_sum(
    inp: TensorType, dims: Optional[Sequence[int]], output_dtype: Optional[int] = None
) -> TensorType:
    """sum(Tensor inp, int[]? dims, *, ScalarType? output_dtype=None) -> Tensor"""

    result = op.ReduceSum(inp, dims, keepdims=False)

    if output_dtype is not None and output_dtype != -1:
        result = op.Cast(result, to=output_dtype)

    return result


def prims_svd(A: TensorType, full_matrices: bool) -> tuple[TensorType, TensorType, TensorType]:
    """svd(Tensor A, *, bool full_matrices) -> (Tensor U, Tensor S, Tensor Vh)"""

    raise NotImplementedError()


@torch_op("prims::tan", trace_only=True)
def prims_tan(self: TTensor) -> TTensor:
    """tan(Tensor self) -> Tensor"""

    return op.Tan(self)


@torch_op("prims::tanh", trace_only=True)
def prims_tanh(self: TTensor) -> TTensor:
    """tanh(Tensor self) -> Tensor"""

    return op.Tanh(self)


@torch_op("prims::transpose", trace_only=True)
def prims_transpose(a: TensorType, permutation: Sequence[int]) -> TensorType:
    """transpose(Tensor(a) a, int[] permutation) -> Tensor(a)"""

    return op.Transpose(a, perm=permutation)


def prims_trunc(self: TensorType) -> TensorType:
    """trunc(Tensor self) -> Tensor"""

    raise NotImplementedError()


def prims_uniform(
    shape: INT64, low: float, high: float, dtype: int, device: str
) -> TensorType:
    """uniform(SymInt[] shape, *, Scalar low, Scalar high, ScalarType dtype, Device device) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::var", trace_only=True)
def prims_var(
    inp: TensorType,
    dims: Optional[Sequence[int]],
    correction: int,
    output_dtype: Optional[int] = None,
) -> TensorType:
    """var(Tensor inp, int[]? dims, *, int correction, ScalarType? output_dtype=None) -> Tensor"""

    if not dims:
        # dims can be empty in practice. We just use a None so it is not added in the ONNX graph
        dims = None
    sub_mean = op.Sub(inp, op.ReduceMean(inp, dims, keepdims=True))
    sqr_mean = op.Mul(sub_mean, sub_mean)
    var = op.ReduceMean(sqr_mean, dims, keepdims=False)
    # Adjust var according to correction value
    if correction != 0:
        inp_shape = op.Shape(inp)
        dim_size = op.Gather(inp_shape, dims, axis=0)
        numel_float = op.CastLike(op.ReduceProd(dim_size, keepdims=False), inp)
        mul = op.Mul(var, numel_float)
        # Subtract the correction value
        sub = op.Sub(numel_float, op.CastLike(correction, inp))
        var = op.Div(mul, sub)

    if output_dtype is not None and output_dtype != -1:
        var = op.Cast(var, to=output_dtype)

    return var


def prims_view_of(a: TensorType) -> TensorType:
    """view_of(Tensor(a) a) -> Tensor"""

    raise NotImplementedError()


@torch_op("prims::where", trace_only=True)
def prims_where(pred: BOOL, a: TTensor, b: TTensor) -> TTensor:
    """where(Tensor pred, Tensor a, Tensor b) -> Tensor"""

    return op.Where(pred, a, b)


def prims_zeta(self: TensorType, other: TensorType) -> TensorType:
    """zeta(Tensor self, Tensor other) -> Tensor"""

    raise NotImplementedError()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

import numpy as np
import onnx.parser

from onnxscript import ir
from onnxscript.rewriter import testing
from onnxscript.rewriter.rules.common import _collapse_slices

_INT64_MAX = 9223372036854775807


class TwoReshapesMatMulReshapeTest(unittest.TestCase):
    def test_slice_is_redundant_when_ends_is_greater_than_input_shape(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[512, 16, 112] data) => (float[512, 16, 112] output)
            {
                starts = Constant<value: tensor = int64[1] {0}>()
                ends = Constant<value: tensor = int64[1] {9999}>()
                axes = Constant<value: tensor = int64[1] {0}>()
                steps = Constant<value: tensor = int64[1] {1}>()
                output = Slice (data, starts, ends, axes, steps)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _collapse_slices.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 5)
        self.assertIn("Identity", [node.op_type for node in model.graph])
        testing.assert_numerically_equal(
            model_proto,
            model,
            (np.random.rand(512, 16, 112).astype(np.float32),),
        )

    def test_slice_is_redundant_when_ends_reaches_int64_max(self):
        model_proto = onnx.parser.parse_model(
            f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[512, 16, 112] data) => (float[512, 16, 112] output)
            {{
                starts = Constant<value: tensor = int64[1] {{0}}>()
                ends = Constant<value: tensor = int64[1] {{{_INT64_MAX}}}>()
                axes = Constant<value: tensor = int64[1] {{0}}>()
                steps = Constant<value: tensor = int64[1] {{1}}>()
                output = Slice (data, starts, ends, axes, steps)
            }}
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _collapse_slices.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 5)
        self.assertIn("Identity", [node.op_type for node in model.graph])
        testing.assert_numerically_equal(
            model_proto,
            model,
            (np.random.rand(512, 16, 112).astype(np.float32),),
        )

    def test_slice_unequal_dynamic_shape(self):
        model_proto = onnx.parser.parse_model(
            f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[L, M, N] data) => (float[P, M, N] output)
            {{
                starts = Constant<value: tensor = int64[1] {{0}}>()
                ends = Constant<value: tensor = int64[1] {{{9}}}>()
                axes = Constant<value: tensor = int64[1] {{0}}>()
                steps = Constant<value: tensor = int64[1] {{1}}>()
                output = Slice (data, starts, ends, axes, steps)
            }}
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _collapse_slices.rules.apply_to_model(model)
        self.assertEqual(count, 0)

    def test_slice_equal_dynamic_shape(self):
        model_proto = onnx.parser.parse_model(
            f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[L, M, N] data) => (float[L, M, N] output)
            {{
                starts = Constant<value: tensor = int64[1] {{0}}>()
                ends = Constant<value: tensor = int64[1] {{{9}}}>()
                axes = Constant<value: tensor = int64[1] {{0}}>()
                steps = Constant<value: tensor = int64[1] {{1}}>()
                output = Slice (data, starts, ends, axes, steps)
            }}
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _collapse_slices.rules.apply_to_model(model)
        self.assertEqual(count, 1)

    def test_slice_equal_dynamic_shape_but_step_reverse(self):
        model_proto = onnx.parser.parse_model(
            f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[L, M, N] data) => (float[L, M, N] output)
            {{
                starts = Constant<value: tensor = int64[1] {{0}}>()
                ends = Constant<value: tensor = int64[1] {{{9}}}>()
                axes = Constant<value: tensor = int64[1] {{0}}>()
                steps = Constant<value: tensor = int64[1] {{-1}}>()
                output = Slice (data, starts, ends, axes, steps)
            }}
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _collapse_slices.rules.apply_to_model(model)
        # Should not change the output shape if we did not use the default step of 1
        self.assertEqual(count, 0)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Function for manipulating input parameters of an Op or a OnnxFunction."""

from __future__ import annotations

import collections
from typing import Any, OrderedDict

from onnxscript import ir


def separate_input_attributes_from_arguments(
    op_signature: ir.schemas.OpSignature,
    args,
    kwargs,
    fill_defaults: bool = True,
    allow_extra_kwargs: bool = False,
    allow_extra_args: bool = True,
) -> tuple[list[Any], OrderedDict[str, Any]]:
    """Separate Python args and kwargs into ONNX inputs and attributes.

    Args:
        op_signature: The operator signature containing parameter information.
        args: The Python positional arguments supplied by the caller.
        kwargs: The Python keyword arguments supplied by the caller.
        fill_defaults: Whether to fill the default values for attributes.
        allow_extra_kwargs: Whether to allow extra keyword arguments.
            When set to True, extra/unknown arguments will be ignored.
        allow_extra_args: Whether to allow extra positional arguments beyond
            what the schema declares (when no variadic parameter exists).
            When set to False, a TypeError is raised for extra args.

    Returns:
        A tuple of two elements:
        - A list of ONNX inputs.
        - An ordered dictionary of ONNX attribute names and values.

    Raises:
        TypeError: When allow_extra_kwargs is False and there are unknown kwargs.
        TypeError: When allow_extra_args is False and there are extra positional args.
        TypeError: When a required input is not provided.
    """
    # args, kwargs and op_signature.params should be all in order
    # user may not specify all inputs or attributes

    all_param_names = {param.name for param in op_signature.params}
    extra_kwargs = set(kwargs).difference(all_param_names)
    if extra_kwargs and not allow_extra_kwargs:
        raise TypeError(f"Unexpected keyword arguments '{extra_kwargs}'")

    onnx_inputs = []
    onnx_attributes = collections.OrderedDict()
    has_variadic = False

    for i, param in enumerate(op_signature.params):
        is_input = param.is_param()
        is_variadic = is_input and param.variadic

        if is_variadic:
            has_variadic = True
            # Exhaust all remaining args
            onnx_inputs.extend(args[i:])
            args = []
            continue
        if i < len(args):
            if is_input:
                onnx_inputs.append(args[i])
            else:
                onnx_attributes[param.name] = args[i]
        elif param.name in kwargs:
            if is_input:
                onnx_inputs.append(kwargs[param.name])
            else:
                onnx_attributes[param.name] = kwargs[param.name]
        elif isinstance(param, ir.schemas.AttributeParameter) and param.has_default():
            # User did not provide the attribute
            if fill_defaults:
                # Extract the value from the Attr object
                onnx_attributes[param.name] = param.default.value
        elif param.required:
            raise TypeError(f"Required input '{param}' was not provided")

    if not allow_extra_args and not has_variadic and len(args) > len(op_signature.params):
        raise TypeError(
            f"Too many positional arguments: expected {len(op_signature.params)}, "
            f"got {len(args)}"
        )

    return onnx_inputs, onnx_attributes


def tag_arguments_with_signature(
    op_signature: ir.schemas.OpSignature,
    args,
    kwargs,
    fill_defaults: bool = True,
    allow_extra_kwargs: bool = False,
) -> tuple[
    list[tuple[Any, ir.schemas.Parameter | ir.schemas.AttributeParameter]],
    dict[str, tuple[Any, ir.schemas.Parameter | ir.schemas.AttributeParameter]],
]:
    """Tag Python args and kwargs with matching ONNX Parameter/AttributeParameter.

    Args:
        op_signature: The operator signature containing parameter information.
        args: The Python positional arguments supplied by the caller.
        kwargs: The Python keyword arguments supplied by the caller.
        fill_defaults: Whether to fill the default values for attributes.
        allow_extra_kwargs: Whether to allow extra keyword arguments.
            When set to True, extra/unknown arguments will be ignored.

    Returns:
        A tuple of two elements:
        - A list of tuple of Python positional argument and Parameter/AttributeParameter.
        - An ordered dictionary of Python keyword argument names and tuple of argument
            value and Parameter/AttributeParameter.

    Raises:
        TypeError: When allow_extra_kwargs is False and there are unknown kwargs.
        TypeError: When a required input is not provided.
    """
    # args, kwargs and op_signature.params should be all in order
    # user may not specify all inputs or attributes

    all_param_names = {param.name for param in op_signature.params}
    extra_kwargs = set(kwargs).difference(all_param_names)
    if extra_kwargs and not allow_extra_kwargs:
        raise TypeError(f"Unexpected keyword arguments '{extra_kwargs}'")

    tagged_args: list[tuple[Any, ir.schemas.Parameter | ir.schemas.AttributeParameter]] = []
    tagged_kwargs: dict[
        str, tuple[Any, ir.schemas.Parameter | ir.schemas.AttributeParameter]
    ] = {}

    for i, param in enumerate(op_signature.params):
        is_variadic = param.is_param() and param.variadic

        if is_variadic:
            # Exhaust all remaining args
            tagged_args.extend((arg, param) for arg in args[i:])
            args = []
            continue
        if i < len(args):
            tagged_args.append((args[i], param))
        elif param.name in kwargs:
            tagged_kwargs[param.name] = (kwargs[param.name], param)
        elif param.has_default():
            # User did not provide the input/attribute
            if fill_defaults:
                default_value = param.default
                # Extract value from Attr object if it's an AttributeParameter
                if param.is_attribute():
                    default_value = param.default.value
                tagged_kwargs[param.name] = (default_value, param)
        elif param.required:
            raise TypeError(f"Required input/attribute '{param}' was not provided")

    return tagged_args, tagged_kwargs


def turn_to_kwargs_to_avoid_ordering(
    op_signature: ir.schemas.OpSignature,
    inputs: list[Any],
    attributes: dict[str, Any],
) -> dict[str, Any]:
    """Return the inputs and attributes to the order of the function signature."""
    for idx, param in enumerate(op_signature.params):
        if param.name not in attributes:
            is_variadic = isinstance(param, ir.schemas.Parameter) and param.variadic
            if is_variadic:
                attributes[param.name] = inputs[idx:]
            elif inputs:
                attributes[param.name] = inputs.pop(0)
    return attributes



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import ast
from typing import Any, Optional, Sequence, Set

from onnxscript._internal import ast_utils, sourceinfo


def _get_loop_var(for_stmt: ast.For, formatter: sourceinfo.Formatter) -> str:
    if not isinstance(for_stmt.target, ast.Name):
        raise TypeError(formatter(for_stmt, "For loop target must be a single variable."))
    return for_stmt.target.id


def _used_vars(expr: Optional[ast.expr]) -> Set[str]:
    """Return set of all variables used, including function names, in an expression."""
    if expr is None:
        return set()
    if isinstance(expr, ast.Name):
        return {expr.id}
    result = set()
    if isinstance(expr, ast.Call):
        # The callee-expression is not visited
        children = expr.args
        for keyword in expr.keywords:
            if isinstance(keyword.value, ast.Name):
                result.add(keyword.value.id)
    else:
        children = ast.iter_child_nodes(expr)  # type: ignore[assignment]
    for c in children:
        result = result | _used_vars(c)
    return result


def _lhs_vars(lhs: ast.expr) -> Set[str]:
    """Return set of assigned variables in the lhs of an assignment statement."""

    def get_id(e):
        assert isinstance(e, ast.Name), "Only simple assignments supported."
        return e.id

    if isinstance(lhs, ast.Tuple):
        return {get_id(x) for x in lhs.elts}
    return {get_id(lhs)}


class AstAnalyzer:
    def __init__(
        self,
        fun: ast.FunctionDef,
        formatter: sourceinfo.Formatter,
        globals: dict[str, Any] | None = None,
    ) -> None:
        self._formatter = formatter
        self._constant_if_condition: dict[ast.If, bool] = {}
        self._live_in: dict[ast.stmt, Set[str]] = {}
        self._live_out: dict[ast.stmt, Set[str]] = {}
        if globals:
            self._compute_constant_if_conditions(fun, globals)
        self.do_liveness_analysis(fun)

    def live_in(self, stmt: ast.stmt) -> Set[str] | None:
        """Get the set of variables that are live at the entry of the given statement."""
        return self._live_in.get(stmt)

    def live_out(self, stmt: ast.stmt) -> Set[str] | None:
        """Get the set of variables that are live at the exit of the given statement."""
        return self._live_out.get(stmt)

    def _compute_constant_if_conditions(
        self, fun: ast.FunctionDef, globals: dict[str, Any]
    ) -> None:
        """Identify if-statements with constant conditions.

        If-statements of the form `if name:` where `name` is an outer-scope variable
        and name is not assigned to within the function body, are treated as constant
        conditions. The value of such conditions is determined from the outer-scope.
        """

        assigned_vars = self.assigned_vars(fun.body)
        for node in ast.walk(fun):
            if isinstance(node, ast.If):
                if isinstance(node.test, ast.Name):
                    python_var = node.test.id
                    if python_var not in assigned_vars and python_var in globals:
                        # Condition depends on an outer-scope variable.
                        self._constant_if_condition[node] = bool(globals[python_var])

    def constant_if_condition(self, if_stmt: ast.If) -> Optional[bool]:
        """Return the constant value of the if-statement condition, if it is constant.

        Args:
            if_stmt: The if-statement-ast to analyze.

        Returns:
            The constant boolean value of the if-statement condition, or None if not constant.
        """
        return self._constant_if_condition.get(if_stmt, None)

    def assigned_vars(self, stmt: ast.stmt | list[ast.stmt]) -> Set[str]:
        """Return the set of all variables that may be assigned to in an execution of input stmt
        or sequence of statements.
        """

        def assigned_in_block(block: Sequence[ast.stmt]) -> Set[str]:
            result: set[Any] = set()
            for s in block:
                result = result | self.assigned_vars(s)
            return result

        if isinstance(stmt, ast.Assign):
            return _lhs_vars(stmt.targets[0])
        if isinstance(stmt, ast.AnnAssign):
            return _lhs_vars(stmt.target)
        if isinstance(stmt, ast.Return):
            return set()
        if isinstance(stmt, ast.If):
            constant_cond = self.constant_if_condition(stmt)
            if constant_cond is None:
                return assigned_in_block(stmt.body) | assigned_in_block(stmt.orelse)
            elif constant_cond:
                return assigned_in_block(stmt.body)
            else:
                return assigned_in_block(stmt.orelse)
        if isinstance(stmt, ast.For):
            return assigned_in_block(stmt.body) | {_get_loop_var(stmt, self._formatter)}
        if isinstance(stmt, ast.While):
            return assigned_in_block(stmt.body)
        if isinstance(stmt, list):
            return assigned_in_block(stmt)
        if isinstance(stmt, ast.Break):
            return set()
        if isinstance(stmt, ast.FunctionDef):
            # Supported function-definitions (used for higher order ops like Scan)
            # do not assign to any variable in the outer scope.
            return set()
        if ast_utils.is_print_call(stmt):
            return set()
        if ast_utils.is_doc_string(stmt):
            return set()
        error_message = self._formatter(stmt, f"Unsupported statement type {type(stmt)!r}.")
        raise ValueError(error_message)

    def do_liveness_analysis(self, fun: ast.FunctionDef) -> None:
        """Perform liveness analysis of the given function-ast."""

        def visit(stmt: ast.stmt, live_out: Set[str]) -> Set[str]:
            self._live_out[stmt] = live_out
            live = do_visit(stmt, live_out)
            self._live_in[stmt] = live
            return live

        def do_visit(stmt: ast.stmt, live_out: Set[str]) -> Set[str]:
            def visit_block(block: Sequence[ast.stmt], live_out: Set[str]) -> Set[str]:
                for s in reversed(block):
                    live_out = visit(s, live_out)
                return live_out

            if isinstance(stmt, ast.Assign):
                return live_out.difference(_lhs_vars(stmt.targets[0])) | _used_vars(stmt.value)
            if isinstance(stmt, ast.AnnAssign):
                return live_out.difference(_lhs_vars(stmt.target)) | _used_vars(stmt.value)
            if isinstance(stmt, ast.Return):
                return _used_vars(stmt.value)
            if isinstance(stmt, ast.If):
                constant_cond = self.constant_if_condition(stmt)
                if constant_cond is None:
                    live1 = visit_block(stmt.body, live_out)
                    live2 = visit_block(stmt.orelse, live_out)
                    return live1 | live2 | _used_vars(stmt.test)
                elif constant_cond:
                    return visit_block(stmt.body, live_out)
                else:
                    return visit_block(stmt.orelse, live_out)
            if isinstance(stmt, ast.For):
                p_loop_var = _get_loop_var(stmt, self._formatter)
                prev = None
                curr = live_out
                while curr != prev:
                    prev = curr
                    curr = visit_block(stmt.body, prev).difference({p_loop_var})
                return curr
            if isinstance(stmt, ast.While):
                cond_vars = _used_vars(stmt.test)
                prev = None
                curr = live_out | cond_vars
                while curr != prev:
                    prev = curr
                    curr = visit_block(stmt.body, prev) | cond_vars
                return curr
            if isinstance(stmt, ast.Break):
                # The following is sufficient for the current restricted usage, where
                # a (conditional) break is allowed only as the last statement of a loop.
                # Break statements in the middle of the loop, however, will require
                # a generalization.
                return live_out
            if ast_utils.is_doc_string(stmt):
                return live_out
            if isinstance(stmt, ast.FunctionDef):
                return live_out
            if ast_utils.is_print_call(stmt):
                return live_out
            raise ValueError(
                self._formatter(stmt, f"Unsupported statement type {type(stmt)!r}.")
            )

        assert isinstance(fun, ast.FunctionDef)
        live: set[Any] = set()
        for s in reversed(fun.body):
            live = visit(s, live)

    def exposed_uses(self, stmts: Sequence[ast.stmt]) -> set[str]:
        """Return the set of variables that are used before being defined by given block.
        In essence, this identifies the "inputs" to a given code-block.
        For example, consider the following code-block:
        ::

        x = x + 10
        y = 20
        z = x + y
        x = 30

        The exposed_uses of this code-block is { x }. The value of z is not used within
        the block. Even though the value of y is used within the block, it is assigned
        a value before it is used. However, in contrast, the incoming value of x is used
        (in the first statement). Hence x is included in the exposed_uses.
        """

        def visit_block(block: Sequence[ast.stmt], live_out: Set[str]) -> Set[str]:
            for stmt in reversed(block):
                live_out = visit(stmt, live_out)
            return live_out

        def visit(stmt: ast.stmt, live_out: Set[str]) -> Set[str]:
            if isinstance(stmt, ast.Assign):
                return live_out.difference(_lhs_vars(stmt.targets[0])) | _used_vars(stmt.value)
            if isinstance(stmt, ast.AnnAssign):
                return live_out.difference(_lhs_vars(stmt.target)) | _used_vars(stmt.value)
            if isinstance(stmt, ast.Return):
                return _used_vars(stmt.value)
            if isinstance(stmt, ast.If):
                constant_cond = self.constant_if_condition(stmt)
                if constant_cond is None:
                    live1 = visit_block(stmt.body, live_out)
                    live2 = visit_block(stmt.orelse, live_out)
                    return (live1 | live2) | _used_vars(stmt.test)
                elif constant_cond:
                    return visit_block(stmt.body, live_out)
                else:
                    return visit_block(stmt.orelse, live_out)
            if ast_utils.is_print_call(stmt):
                return live_out
            if ast_utils.is_doc_string(stmt):
                return live_out
            if isinstance(stmt, ast.For):
                # Analysis assumes loop may execute zero times. Results can be improved
                # for loops that execute at least once.
                loop_var_set = {_get_loop_var(stmt, self._formatter)}
                used_after_loop = live_out.difference(loop_var_set)
                used_inside_loop = visit_block(stmt.body, set()).difference(loop_var_set)
                used_in_loop_header = _used_vars(stmt.iter)
                return used_inside_loop | used_in_loop_header | used_after_loop
            if isinstance(stmt, ast.While):
                # Analysis assumes loop may execute zero times. Results can be improved
                # for loops that execute at least once.
                used_inside_loop = visit_block(stmt.body, set())
                used_in_loop_header = _used_vars(stmt.test)
                return used_inside_loop | used_in_loop_header | live_out
            if isinstance(stmt, ast.Break):
                # Currently, we assume that break statements are only allowed as the last
                # statement in a loop, as "if cond: break".
                return live_out
            if isinstance(stmt, ast.FunctionDef):
                if stmt.name in live_out:
                    live_out.remove(stmt.name)
                    live_out = live_out | self.outer_scope_variables(stmt)
                return live_out
            raise ValueError(
                self._formatter(stmt, f"Unsupported statement type {type(stmt)!r}.")
            )

        return visit_block(stmts, set())

    def outer_scope_variables(self, fun: ast.FunctionDef) -> set[str]:
        """Return the set of outer-scope variables used in a nested function.

        Args:
            fun: The function-ast to analyze.
            formatter: The formatter object.

        Returns:
            A set of variable names (strings).
        """
        assert isinstance(fun, ast.FunctionDef)
        used_vars_ = self.exposed_uses(fun.body)
        inputs = [x.arg for x in fun.args.args]
        return used_vars_.difference(inputs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=import-outside-toplevel
from __future__ import annotations

from typing import Any, Sequence

import torch

import onnxscript.tools.transformers_models


def _prepare_config_and_inputs(
    batch_size: int,
    seq_length: int,
    vocab_size: int,
    type_sequence_label_size: int = 2,
    type_vocab_size: int = 16,
    num_labels: int = 3,
    num_choices: int = 4,
    use_input_mask: bool = False,
    use_token_type_ids: bool = False,
    use_labels: bool = False,
) -> tuple[Any, ...]:
    input_ids = onnxscript.tools.transformers_models.ids_tensor(
        [batch_size, seq_length], vocab_size
    )

    input_mask = None
    if use_input_mask:
        input_mask = torch.tril(torch.ones(batch_size, seq_length))

    token_type_ids = None
    if use_token_type_ids:
        assert type_vocab_size > 0, "type_vocab_size is null"
        token_type_ids = onnxscript.tools.transformers_models.ids_tensor(
            [batch_size, seq_length], type_vocab_size
        )

    sequence_labels = None
    token_labels = None
    choice_labels = None
    if use_labels:
        assert type_sequence_label_size > 0, "type_sequence_label_size is null"
        assert num_labels > 0, "num_labels is null"
        assert num_choices > 0, "num_choices is null"
        sequence_labels = onnxscript.tools.transformers_models.ids_tensor(
            [batch_size], type_sequence_label_size
        )
        token_labels = onnxscript.tools.transformers_models.ids_tensor(
            [batch_size, seq_length], num_labels
        )
        choice_labels = onnxscript.tools.transformers_models.ids_tensor(
            [batch_size], num_choices
        )

    return (
        input_ids,
        token_type_ids,
        input_mask,
        sequence_labels,
        token_labels,
        choice_labels,
    )


def get_phi_model(
    input_dims: Sequence[tuple[int, int]] = ((13, 7), (14, 7), (15, 8)),
    hidden_size: int = 32,
    num_hidden_layers: int = 2,
    vocab_size: int = 99,
    intermediate_size: int = 16,
    max_position_embeddings: int = 512,
    num_attention_heads: int = 4,
    num_key_value_heads: int = 2,
    _attn_implementation: str = "eager",  # needed value to remove graph breaks
    with_mask: bool = True,
) -> tuple[Any, list[tuple[torch.Tensor, ...]], dict]:
    """
    Returns a model.
    See `PhiConfig
    <https://huggingface.co/docs/transformers/main/en/model_doc/phi#transformers.PhiConfig>`_.
    The parameters are chosen for a unit test configuration from `test_modeling_phi.py
    <https://github.com/huggingface/transformers/blob/main/tests/models/phi/test_modeling_phi.py>`_.
    """
    from transformers import PhiConfig
    from transformers.models.phi.modeling_phi import PhiModel

    dynamic_shapes = {0: {0: "batch", 1: "length"}}
    if with_mask:
        dynamic_shapes.update({1: {0: "batch", 1: "length"}})

    config = PhiConfig(
        hidden_size=hidden_size,
        num_hidden_layers=num_hidden_layers,
        vocab_size=vocab_size,
        intermediate_size=intermediate_size,
        max_position_embeddings=max_position_embeddings,
        num_attention_heads=num_attention_heads,
        num_key_value_heads=num_key_value_heads,
    )
    if _attn_implementation:
        config._attn_implementation = _attn_implementation  # pylint: disable=protected-access

    if with_mask:

        class PhiModelWrapper(torch.nn.Module):
            def __init__(self, config):
                super().__init__()
                self.model = PhiModel(config)

            def forward(self, input_ids, attention_mask):
                model_output = self.model(
                    input_ids, attention_mask=attention_mask, use_cache=False
                )
                return model_output.to_tuple()

        def generate_example_inputs(batch: int, seq: int, vocab_size: int):
            (
                input_ids,
                _,  # token_type_ids,
                input_mask,
                _,  # sequence_labels,
                _,  # token_labels,
                _,  # choice_labels,
            ) = _prepare_config_and_inputs(
                batch_size=batch,
                seq_length=seq,
                vocab_size=vocab_size,
                use_input_mask=True,
            )
            return input_ids, input_mask

        example_args_collection = []
        for b, s in input_dims:
            example_args_collection.append(generate_example_inputs(b, s, vocab_size))

        return PhiModelWrapper(config), example_args_collection, dynamic_shapes

    # no mask

    class PhiModelWrapperNoMask(torch.nn.Module):
        def __init__(self, config):
            super().__init__()
            self.model = PhiModel(config)

        def forward(self, input_ids):
            model_output = self.model(input_ids, use_cache=False)
            return model_output.to_tuple()

    def generate_example_inputs_no_mask(batch: int, seq: int, vocab_size: int):
        (
            input_ids,
            _,  # token_type_ids,
            _,  # input_mask,
            _,  # sequence_labels,
            _,  # token_labels,
            _,  # choice_labels,
        ) = _prepare_config_and_inputs(
            batch_size=batch,
            seq_length=seq,
            vocab_size=vocab_size,
            use_input_mask=True,
        )
        return (input_ids,)

    example_args_collection = []
    for b, s in input_dims:
        example_args_collection.append(generate_example_inputs_no_mask(b, s, vocab_size))

    return PhiModelWrapperNoMask(config), example_args_collection, dynamic_shapes


def get_phi_model_from_config(
    warmup: int = 5,
    repeat: int = 10,
    config: str = "small",
    num_hidden_layers: int = 1,
    implementation: str = "eager",
    dynamic_shapes: bool = False,
    with_mask: bool = True,
) -> tuple[Any, list[tuple[torch.Tensor, ...]], dict]:
    """
    Returns a model Phi to test or benchmark.

    Args:
        warmup: Number of inputs to generate.
        repeat: Number of inputs to generate for repeat.
        config: small, medium or large
        num_hidden_layers: number of hidden layers
        implementation: eager or sdpa
        with_mask: One or two inputs.
        dynamic_shapes: dynamic shapes or not

    Returns:
        Model and list of inputs.
    """
    if config == "small":
        conf_dict = dict(
            input_dims=onnxscript.tools.transformers_models.get_input_dims_for_llm(
                dynamic_shapes, warmup, repeat
            ),
            hidden_size=32,
            num_hidden_layers=num_hidden_layers,
            vocab_size=99,
            intermediate_size=16,
            max_position_embeddings=512,
            num_attention_heads=4,
            num_key_value_heads=2,
            _attn_implementation=implementation,
            with_mask=with_mask,
        )
    elif config == "medium":
        conf_dict = dict(
            input_dims=onnxscript.tools.transformers_models.get_input_dims_for_llm(
                dynamic_shapes, warmup, repeat
            ),
            hidden_size=1024,
            num_hidden_layers=num_hidden_layers,
            vocab_size=1024,
            intermediate_size=1024,
            num_attention_heads=4,
            num_key_value_heads=4,
            max_position_embeddings=1024,
            _attn_implementation=implementation,
            with_mask=with_mask,
        )
    elif config in ("large", "default"):
        conf_dict = dict(
            input_dims=onnxscript.tools.transformers_models.get_input_dims_for_llm(
                dynamic_shapes, warmup, repeat
            ),
            hidden_size=2048,
            num_hidden_layers=num_hidden_layers,
            vocab_size=51200,
            intermediate_size=8192,
            num_attention_heads=32,
            num_key_value_heads=None,
            max_position_embeddings=2048,
            _attn_implementation=implementation,
            with_mask=with_mask,
        )
    else:
        raise ValueError(f"Unexpected configuration {config!r}.")

    return get_phi_model(**conf_dict)  # type: ignore[arg-type]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

__all__ = [
    "ConvertVersionPass",
    "convert_version",
]

import logging

import onnx
import onnx_ir.passes.common as common_passes

from onnxscript import ir
from onnxscript.version_converter import _c_api_utils, _version_converter

logger = logging.getLogger(__name__)


class ConvertVersionPass(ir.passes.InPlacePass):
    """Convert the model to the specified ONNX opset version.

    This pass leverages the onnxscript version converter to convert the model. If
    the conversion is not supported, it falls back to the onnx C API to convert
    the model. This pass is in-place.

    The pass is an no-op if the c-api fails.

    Attributes:
        target_version: The target ONNX opset version to convert the model to.
        fallback: Whether to fallback to the onnx version converter if the
            target version is not supported. Default is False.
    """

    def __init__(self, target_version: int, fallback: bool = False) -> None:
        super().__init__()
        self.target_version = target_version
        self.fallback = fallback
        self._inline_pass = ir.passes.Sequential(
            # NOTE: The current version converter only supports inlined models.
            common_passes.InlinePass(),
            # NOTE: Old torch version might include legacy Rank onnx functions which is
            # not used in the converted model.
            common_passes.RemoveUnusedFunctionsPass(),
            common_passes.RemoveUnusedOpsetsPass(),
        )
        self._convert_pass = _ConvertVersionPassRequiresInline(
            target_version=target_version,
            fallback=fallback,
        )
        self._cleanup_passes = ir.passes.Sequential(
            common_passes.RemoveUnusedNodesPass(),
            common_passes.RemoveUnusedFunctionsPass(),
            common_passes.RemoveUnusedOpsetsPass(),
        )

    def call(self, model: ir.Model) -> ir.passes.PassResult:
        # Run the conversion pass outside of Sequential so that errors
        # (e.g. VersionConverterError) propagate directly without being
        # wrapped in PassError.
        inline_result = self._inline_pass(model)
        result = self._convert_pass(inline_result.model)
        cleanup_result = self._cleanup_passes(result.model)
        return ir.passes.PassResult(
            cleanup_result.model,
            result.modified or cleanup_result.modified or inline_result.modified,
        )


class _ConvertVersionPassRequiresInline(ir.passes.InPlacePass):
    """Convert the model to the specified ONNX opset version.

    This pass leverages the onnxscript version converter to convert the model. If
    the conversion is not supported, it falls back to the onnx C API to convert
    the model. This pass is in-place.

    The pass is an no-op if the c-api fails.

    Attributes:
        target_version: The target ONNX opset version to convert the model to.
        fallback: Whether to fallback to the onnx version converter if the
            target version is not supported.
    """

    def __init__(self, target_version: int, fallback: bool) -> None:
        super().__init__()
        self.target_version = target_version
        self.fallback = fallback

    def call(self, model: ir.Model) -> ir.passes.PassResult:
        if "" in model.graph.opset_imports:
            onnx_opset_version = model.graph.opset_imports[""]
            if onnx_opset_version == self.target_version:
                # No need to convert the version
                return ir.passes.PassResult(model, False)

        # When fallback is disabled, always use the onnxscript version converter;
        # When fallback is enabled, use the onnxscript version converter
        # if the target version is supported. Otherwise, use the onnx C API
        # to convert the model.
        if not self.fallback or _version_converter.version_supported(
            model, self.target_version
        ):
            _version_converter.convert_version(
                model,
                target_version=self.target_version,
            )
            return ir.passes.PassResult(model, True)

        if not self.fallback:
            logger.warning(
                "The model version conversion is not supported by the onnxscript version converter "
                "and fallback is disabled. The model was not modified"
                " (target version: %d). "
                "Set fallback=True to enable fallback to the onnx c-api version converter.",
                self.target_version,
            )
            return ir.passes.PassResult(model, False)
        else:
            logger.warning(
                "The model version conversion is not supported by the onnxscript version converter "
                "and fallback is enabled. The model will be converted using the onnx C API "
                "(target version: %d).",
                self.target_version,
            )

        # If the onnxscript version converter does not support the conversion,
        # we can use the onnx C API to convert the model
        def _partial_convert_version(proto: onnx.ModelProto) -> onnx.ModelProto:
            """Partial function to check the model."""
            return onnx.version_converter.convert_version(
                proto, target_version=self.target_version
            )

        try:
            converted_proto = _c_api_utils.call_onnx_api(
                func=_partial_convert_version, model=model
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.warning(
                "Failed to convert the model to the target version %d using the ONNX C API. "
                "The model was not modified",
                self.target_version,
                exc_info=e,
            )
            return ir.passes.PassResult(model, False)

        converted_model = ir.from_proto(converted_proto)

        # Recover the initializers in the converted model
        for input in converted_model.graph.inputs:
            if input.name in model.graph.initializers:
                input.const_value = model.graph.initializers[input.name].const_value
                converted_model.graph.register_initializer(input)
        user_inputs = converted_model.graph.inputs[: len(model.graph.inputs)]
        converted_model.graph.inputs.clear()
        converted_model.graph.inputs.extend(user_inputs)

        # Return the converted graph to the original model to keep the pass in-place
        model.graph = converted_model.graph
        return ir.passes.PassResult(model, True)


def convert_version(
    model: ir.Model | onnx.ModelProto, target_version: int, fallback=None
) -> None:
    """Convert the model to the specified ONNX opset version.

    Args:
        model: The model to convert.
        target_version: The target ONNX opset version.
        fallback: Whether to fallback to the onnx version converter if the
            target version is not supported. Default is False.
    """
    if isinstance(model, onnx.ModelProto):
        model_proto = model
        model = ir.from_proto(model)
    else:
        model_proto = None

    assert isinstance(model, ir.Model)
    ConvertVersionPass(target_version=target_version, fallback=fallback)(model)

    if model_proto is not None:
        # Update the model proto in-place
        model_proto.graph.Clear()
        del model_proto.functions[:]
        model_proto.graph.CopyFrom(ir.to_proto(model.graph))



# --------------------------------------------------------------------------
# ⚠️ WARNING - AUTO-GENERATED CODE - DO NOT EDIT ⚠️
# ⚙️ Generated by 'python -m opgen'
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
# pylint: disable=W0221,W0222,R0901,W0237
# mypy: disable-error-code=override
# ruff: noqa: D402
# --------------------------------------------------------------------------

from __future__ import annotations

from typing import TypeVar

from onnx.defs import get_schema

from onnxscript.onnx_opset._impl.opset3 import Opset3
from onnxscript.onnx_types import (
    BOOL,
    COMPLEX64,
    COMPLEX128,
    DOUBLE,
    FLOAT,
    FLOAT16,
    INT8,
    INT16,
    INT32,
    INT64,
    STRING,
    UINT8,
    UINT16,
    UINT32,
    UINT64,
)
from onnxscript.values import Op, Opset


class Opset4(Opset3):
    def __new__(cls):
        return Opset.__new__(cls, "", 4)

    T_Concat = TypeVar(
        "T_Concat",
        BOOL,
        COMPLEX128,
        COMPLEX64,
        DOUBLE,
        FLOAT,
        FLOAT16,
        INT16,
        INT32,
        INT64,
        INT8,
        STRING,
        UINT16,
        UINT32,
        UINT64,
        UINT8,
    )

    def Concat(self, *inputs: T_Concat, axis: int) -> T_Concat:
        r"""[🌐 Concat(4)](https://onnx.ai/onnx/operators/onnx__Concat.html#concat-4 "Online Documentation")

        Concatenate a list of tensors into a single tensor

        Args:
            inputs: (variadic) List of tensors for concatenation

            axis: Which axis to concat on
        """

        schema = get_schema("Concat", 4, "")
        op = Op(self, "Concat", schema)
        return op(*self._prepare_inputs(schema, *inputs), axis=axis)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import ast
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    NoReturn,
    Sequence,
    Union,
)

import onnx_ir as ir

import onnxscript
from onnxscript import onnx_types
from onnxscript._internal import (
    analysis,
    ast_utils,
    autocast,
    builder,
    irbuilder,
    param_manipulation,
    sourceinfo,
    values,
)
from onnxscript._internal import (
    type_annotation as ta,
)
from onnxscript.ir import _schemas

logger = logging.getLogger("onnxscript")


# Python-to-IR converter:


class TranslationError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def warn(msg):
    logger.warning(msg)


def fail(msg) -> NoReturn:
    raise TranslationError(msg)


def fail_if(cond, msg):
    if cond:
        raise TranslationError(msg)


def ignore(cond, msg):
    if cond:
        warn(msg)


# map from python operators to ONNX ops
primop_map = {
    ast.Add: "Add",
    ast.And: "And",
    ast.BitAnd: "And",
    ast.BitOr: "Or",
    ast.Div: "Div",
    ast.Eq: "Equal",
    ast.Gt: "Greater",
    ast.GtE: "GreaterOrEqual",
    ast.Lt: "Less",
    ast.LtE: "LessOrEqual",
    ast.MatMult: "MatMul",
    ast.Mod: "Mod",
    ast.Mult: "Mul",
    ast.Not: "Not",
    ast.NotEq: "NotEqual",
    ast.Or: "Or",
    ast.Pow: "Pow",
    ast.Sub: "Sub",
    ast.USub: "Neg",
}


if TYPE_CHECKING:
    # The type-alias LocalSymValue represents the types of values that local names in a
    # script-function may be bound to during translation, (ONNX IR values).
    # TODO(rama): Rationalize this and values.SymbolValue

    LocalSymValue = Union[values.SymbolValue, irbuilder.IRFunction]

    # The type-alias PyValue is used to represent the types of python values that may be used
    # in an ONNX Script function.
    # TODO(rama): Flesh out the set of valid types here. These include values such as
    # 1 (int), 1.0 (float), [2, 4], [1.0], etc. which will be converted to ONNX, for
    # use as value-parameters or attribute-parameters in an ONNX call (Node).

    PyValue = Any

    # The type-alias SymValue denotes values that an identifier may be bound to during
    # translation. A local name will be bound to a LocalSymValue, while a global name
    # will be bound to a PyValue.

    SymValue = Union[LocalSymValue, PyValue]

    # PreferredName is a type-alias used to represent the preferred name used in the generated
    # ONNX for a value returned by an expression. There is no guarantee that the specified
    # name will be used exactly. The converter will modify the name (with a suffix),
    # if necesssary, to ensure that it is unique (to ensure ONNX's SSA requirement).

    PreferredName = str

    # The type-alias OnnxVar indicates variable names used in the generated ONNX.
    OnnxVarName = str


def set_type_info(value: ir.Value, typeinfo: ta.TypeAnnotationValue) -> None:
    """Sets the type information on an IR value."""
    try:
        type_and_shape = ir.from_proto(typeinfo.to_type_proto())
        value.type = type_and_shape.type
        value.shape = type_and_shape.shape
    except AttributeError:
        # TODO: This needs to be fixed.
        pass
    value.meta["typeinfo"] = typeinfo


def make_value(
    varname: str, typeinfo: ta.TypeAnnotationValue, source_info: sourceinfo.SourceInfo
) -> ir.Value:
    value = ir.Value(name=varname)
    value.meta.setdefault("sourceinfo", source_info)
    if typeinfo is not None:
        set_type_info(value, typeinfo)
    return value


class Converter:
    """Main class to translate python code into ONNX operators.

    Args:
        ir_builder: convert AST node into ONNX structures, if None,
            class :class:`onnxscript.irbuilder.IRBuilder` is used

    The class uses logger `onnxscript`. Logging can be enabled with the following code:

    ::

        import logging
        logging.basicConfig(level=logging.DEBUG)

    Or if you need to enable only the logger used by this module:

    ::

        import logging
        logger = logging.getLogger('onnxscript')
        logger.setLevel(logging.DEBUG)
        console = logging.StreamHandler()
        logger.addHandler(console)
    """

    def __init__(
        self,
        opset: values.Opset | None = None,
        global_names: dict[str, Any] | None = None,
        source: str | None = None,
        default_opset: Union[values.Opset, builder.OpBuilder, None] = None,
    ):
        self.source = source
        if global_names is not None:
            # We make a copy in case function eval modifies it.
            self.globals = global_names.copy()
        self.this_module = opset
        # Convert OpBuilder to Opset if necessary and store the converted value
        if isinstance(default_opset, builder.OpBuilder):
            self.default_opset_ = values.Opset(default_opset.domain, default_opset.version)
        else:
            self.default_opset_ = default_opset

        # States initialized by `_init_function_translation`
        self._outer: list[irbuilder.IRFunction] = []
        self._current_fn: irbuilder.IRFunction = None
        self._nextvar: int = 0
        self._used_vars: set[str] = set()
        self._locals: list[dict[str, LocalSymValue]] = [{}]
        self._analyzer: analysis.AstAnalyzer | None = None
        self._castable: set[str] = set()

    def _is_castable(self, var_name: str) -> bool:
        """Returns True if the variable with the given name represents a polymorphic constant."""
        return var_name in self._castable

    @property
    def analyzer(self) -> analysis.AstAnalyzer:
        if self._analyzer is None:
            raise RuntimeError("Analyzer not initialized.")
        return self._analyzer

    @property
    def default_opset(self) -> values.Opset:
        if self.default_opset_ is None:
            raise RuntimeError(
                "default_opset must be specified in script for functions "
                "that do not contain any use of an ONNX opset."
            )
        return self.default_opset_

    def _set_default_opset(self, opset: values.Opset, node: ast.AST) -> None:
        if opset.domain != "":
            return
        if self.default_opset_ is not None:
            if (
                opset.domain != self.default_opset_.domain
                or opset.version != self.default_opset_.version
            ):
                self._fail(
                    node, f"Two distincts opset were used ({opset} != {self.default_opset_})."
                )
        else:
            self.default_opset_ = opset

    def _find_onnx_opset(self, node: ast.AST) -> values.Opset | None:
        """Find the (first) ONNX opset used in the function, if any."""
        # Search for a Call expression of form "op.OpName(...)"
        if isinstance(node, ast.Call) and isinstance(node.func, ast.Attribute):
            opset_expr = node.func.value
            if isinstance(opset_expr, ast.Name) and opset_expr.id in self.globals:
                opset = self.globals[opset_expr.id]
                # Accept both values.Opset and builder.OpBuilder
                if isinstance(opset, values.Opset):
                    if opset.domain == "":
                        return opset
                elif isinstance(opset, builder.OpBuilder):
                    if opset.domain == "":
                        return values.Opset(opset.domain, opset.version)
        for child in ast.iter_child_nodes(node):
            res = self._find_onnx_opset(child)
            if res is not None:
                return res
        return None

    def _init_function_translation(self) -> None:
        """Initialize self for translating a new (top-level) function."""
        self._outer = []
        self._current_fn: irbuilder.IRFunction | None = None
        self._nextvar = 0
        self._used_vars = set()
        self._locals: list[dict[str, LocalSymValue]] = [{}]

    def _source_of(self, node: ast.AST) -> sourceinfo.SourceInfo:
        return sourceinfo.SourceInfo(
            node, code=self.source, function_name=self._current_fn.name
        )

    def _message(self, node: ast.AST, error_msg: str) -> str:
        """Constructs an error _message containing source information about an ast node."""
        return self._source_of(node).msg(error_msg)

    def _warn(self, node: ast.AST, error_msg: str) -> None:
        warn(self._message(node, error_msg))

    def _fail(self, node: ast.AST, error_msg: str) -> NoReturn:
        fail(self._message(node, error_msg))

    # Name resolution and namescopes: This component handles the following aspects:
    # * Name-scopes are different in Python and the generated ONNX:
    #   - Control-flow blocks (a loop body or the then-or-else block of an if-stmt)
    #     form part of the same name-scope in Python, but will be mapped to a nested
    #     name-scope (as a sub-graph) in ONNX.
    # * Script-time name-value tracking: Name _lookup during script-time returns
    #   statically-known information about the value the name will have at runtime.
    def _enter_scope(self, name: str, parent_node: ast.AST):
        """Enter a control-flow block (a loop body or if-then-else branch).
        The block is translated into a nested-scope in ONNX.
        """
        self._outer.append(self._current_fn)
        self._current_fn = irbuilder.IRFunction(name)
        self._locals.append({})
        logger.debug("Converter:_enter_scope:%d:node:%s", len(self._locals), type(parent_node))

    def _exit_scope(self) -> irbuilder.IRFunction:
        """Exit from a control-flow block (a loop body or if-then-else branch)."""
        logger.debug("Converter:_exit_scope:%d", len(self._locals))
        graph = self._current_fn
        self._current_fn = self._outer.pop()
        self._locals.pop()
        return graph

    def _current_scope(self) -> dict[str, LocalSymValue]:
        return self._locals[-1]

    def _bind(self, name: str, val: LocalSymValue) -> None:
        logger.debug("Converter:_bind:%s", name)
        self._locals[-1][name] = val

    def _lookup(
        self, name: str, info: sourceinfo.SourceInfo, raise_exception: bool = True
    ) -> SymValue:
        """Maps a python variable name to the corresponding value used during translation.

        Typically, a python variable X will correspond to an ONNX value Y. But other special
        cases include: constant values or functions (mapped to Graph attributes), etc.
        """

        for scope in reversed(self._locals):
            if name in scope:
                return scope[name]
        if name in self.globals:
            return self.globals[name]
        if raise_exception:
            raise ValueError(info.msg(f"Unbound name: {name}."))
        return None

    def _generate_unique_name(self, candidate: str = "tmp") -> str:
        # TODO(justinchuby): Can we reduce the O complexity of this function?
        r = candidate
        while r in self._used_vars:
            r = f"{candidate}_{self._nextvar}"
            self._nextvar = self._nextvar + 1
        self._used_vars.add(r)
        return r

    def _to_onnx_attr_ref(
        self, val: values.AttrRef, info: sourceinfo.SourceInfo | None
    ) -> ir.Attr:
        attrtype = val.value.type
        attrname = None
        if attrtype is ir.AttributeType.FLOAT:  # onnx.AttributeProto.FLOAT:
            attrname = "value_float"
        elif attrtype is ir.AttributeType.INT:
            attrname = "value_int"
        elif attrtype is ir.AttributeType.STRING:
            attrname = "value_string"
        elif attrtype is ir.AttributeType.INTS:
            attrname = "value_ints"
        else:
            msg = f"Unsupported attribute type {attrtype!r}."
            fail(info.msg(msg) if info else msg)
        return ir.Attr(attrname, attrtype, value=None, ref_attr_name=val.value.name)

    def _to_onnx_var(
        self,
        val: values.SymbolValue | PyValue,
        target: PreferredName | None = None,
        info: sourceinfo.SourceInfo | None = None,
    ) -> ir.Value:
        if isinstance(val, values.AttrRef):
            # promote attribute to value
            result_name = self._generate_unique_name(target or "tmp")
            attr = self._to_onnx_attr_ref(val, info)
            result = self._emit(
                [result_name], values.Op(self.default_opset, "Constant"), [], [attr]
            )
            if val.as_bool:
                # ONNX attributes use an int-encoding for bools, but ONNX tensor types
                # distinguish between int and bool. So we cast the int tensor to a bool tensor,
                # to promote a (python) bool attribute to a ONNX bool tensor.
                result_as_bool = self._generate_unique_name(result_name + "_as_bool")
                cast_attr = ir.AttrInt64("to", onnx_types.BOOL.dtype)
                self._castable.add(result_as_bool)
                return self._emit1(
                    [result_as_bool],
                    values.Op(self.default_opset, "Cast"),
                    [result],
                    [cast_attr],
                )
            self._castable.add(result_name)
            return result
        if isinstance(val, values.SymbolValue):
            if isinstance(val.value, ir.Value):
                return val.value
        if isinstance(val, ir.Value):
            # An outer-scope ir.Value (e.g., from a closure variable) can be used directly.
            return val
        # Assume value is a python-value convertible to a tensor
        # TODO: check if value is convertible to a TensorProto, so that we can
        # produce a better error _message otherwise
        return self._emit_const(val, target or "tmp", info)

    def _py_var_to_onnx_var(self, py_var: str, info: sourceinfo.SourceInfo) -> ir.Value:
        return self._to_onnx_var(self._lookup(py_var, info), target=py_var, info=info)

    def _emit(
        self,
        outputs: Sequence[str],
        callee: values.Op | str,
        inputs: Sequence[ir.Value | None],
        attrs: Sequence[ir.Attr] | None = None,
    ) -> Sequence[ir.Value] | ir.Value:
        if not isinstance(callee, values.Op):
            callee = values.Op(self.default_opset, callee)
        if attrs is None:
            attrs = []
        output_values = [ir.Value(name=o) for o in outputs]
        node = ir.Node(
            domain=callee.opset.domain,
            version=callee.opset.version,
            op_type=callee.name,
            inputs=inputs,
            outputs=output_values,
            attributes=attrs,
        )
        if not isinstance(callee, values.Op):
            raise TypeError(f"Unexpected type {type(callee)} for callee.")
        node.meta.setdefault("callee", callee)
        assert self._current_fn is not None
        self._current_fn.append_node(node)

        return output_values if len(output_values) > 1 else output_values[0]

    def _emit1(self, *args, **kwargs) -> ir.Value:
        r = self._emit(*args, **kwargs)
        if not isinstance(r, ir.Value):
            raise TypeError(f"Expected single ONNX IR Value, got {type(r)!r}.")
        return r

    def _emit_const(
        self,
        pyvalue: PyValue,
        suggested_name: PreferredName | None,
        info: sourceinfo.SourceInfo,
    ) -> ir.Value:
        if suggested_name is None:
            if isinstance(pyvalue, int):
                if pyvalue >= 0:
                    suggested_name = f"int64_{pyvalue}"
                else:
                    suggested_name = f"int64_m{abs(pyvalue)}"
            elif (
                isinstance(pyvalue, list) and len(pyvalue) == 1 and isinstance(pyvalue[0], int)
            ):
                if pyvalue[0] >= 0:
                    suggested_name = f"int64_{pyvalue[0]}_1d"
                else:
                    suggested_name = f"int64_m{abs(pyvalue[0])}_1d"
            else:
                suggested_name = "const"
        ovar = self._generate_unique_name(suggested_name)

        try:
            tensor = ir.tensor(pyvalue, name=ovar)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self._fail(
                info.ast_node,
                f"Failed to convert constant value {pyvalue!r} to ONNX tensor: {exc}",
            )
        attr = ir.AttrTensor("value", tensor)
        self._castable.add(ovar)
        return self._emit1([ovar], values.Op(self.default_opset, "Constant"), [], [attr])

    def _emit_copy(self, original_var: ir.Value, suggested_name: str) -> ir.Value:
        """Emits a copy statement, using the ONNX Identity operator."""
        new_var = self._generate_unique_name(suggested_name)
        return self._emit([new_var], "Identity", [original_var])

    def _is_constant_expr(self, node: ast.AST) -> None:
        if isinstance(node, ast.UnaryOp):
            return self._is_constant_expr(node.operand)
        if isinstance(
            node,
            (
                ast.Call,
                ast.BinOp,
                ast.UnaryOp,
                ast.Compare,
                ast.Attribute,
                ast.List,
                ast.Load,
                ast.Constant,
            ),
        ):
            return all(self._is_constant_expr(c) for c in ast.iter_child_nodes(node))
        return False

    def _eval_constant_expr(self, expr: ast.expr) -> PyValue:
        """Evaluates a sub-expression that is assumed to represent a constant value.
        The expression can refer only to global names (inherited from the scope
        where the script is evaluated) and cannot refer to local names defined
        within the script.) Further, these expressions are assumed to be constants.
        Thus, any subsequent mutation of any state/variables (used in computing
        this constant value) will potentially lead to unexpected behavior (such
        as divergence between eager-mode execution and evaluation of the ONNX
        function.)
        """
        # TODO: assert (self._is_constant_expr(expr))
        locals: dict[Any, Any] = {}
        cpl = compile(ast.Expression(expr), filename="<ast>", mode="eval")
        try:
            return eval(cpl, self.globals, locals)  # pylint: disable=eval-used
        except NameError as e:
            raise NameError(
                self._message(
                    expr,
                    f"Missing names, globals contains {list(self.globals)!r}, "
                    f"locals {list(locals)!r}.",
                )
            ) from e

    def _get_type_annotation(self, annotation: ast.expr) -> ta.TypeAnnotationValue | None:
        typeinfo = self._eval_constant_expr(annotation)
        if not ta.is_valid_type(typeinfo):
            self._warn(
                annotation,
                "Unsupported type annotation.",
            )
            typeinfo = None
        return typeinfo

    def _translate_attr(
        self,
        attr_name: str,
        expr: ast.AST,
        attr_meta: ir.schemas.AttributeParameter | None = None,
    ) -> ir.Attr | None:
        """Translate an attribute-value specification of the form `attr_name=<expr>`
        in a call to an op. expr is an AST. The following cases are supported:
        * Expr evaluates to a script-time constant (a python-value) that can be mapped
        into an ONNX attribute value, or
        * Expr evaluates to None, in which case None is returned, or
        * Expr must be an attribute-reference, that is a name representing an
        attribute-parameter of a containing function.
        """

        if isinstance(expr, ast.Name):
            val = self._lookup(expr.id, self._source_of(expr))
            if isinstance(val, values.SymbolValue):
                val = val.value
                if isinstance(val, ir.Attr):
                    # A reference to an attribute parameter:
                    attr = val
                    attr_ref = ir.Attr(
                        attr_name, attr.type, value=None, ref_attr_name=attr.name
                    )
                    if attr_meta is not None and (attr.type != attr_meta.type):
                        self._fail(
                            expr,
                            f"Attribute type '{attr_ref.type}' does not match expected type '{attr_meta.type}'",
                        )
                    return attr_ref
                if isinstance(val, irbuilder.IRFunction):
                    # A reference to a nested-function: convert to GraphProto and use it.
                    irfunction = val
                    # Check that outer-scope variables referenced by function have same value
                    # at function-definition site and use-as-attribute site, to avoid errors.
                    for pyvar, previous in irfunction.outer_scope_variables:
                        current = self._lookup(pyvar, self._source_of(expr))
                        if current.value != previous.value:
                            self._fail(
                                expr,
                                f"Outer scope variable '{pyvar}' referenced by function "
                                f"'{expr.id!r}' modified.",
                            )

                    val = irfunction.graph
                if isinstance(val, ir.Value):
                    self._fail(expr, f"Cannot use ir.Value '{expr.id}' as an attribute.")
                else:
                    # Treat as a constant python-value, to be converted below.
                    pass
            else:
                # This must be a reference to an outer-scope python-value, typically a constant.
                # The value will be converted to an ONNX attribute value below.
                pass
        else:
            val = self._eval_constant_expr(expr)

        # In ONNX, there is no way to explicitly specify a None value for an attribute.
        # Instead, the attribute must be omitted from the attribute list.
        # Hence, we do not create an attribute-proto if the value is None.
        # The caller is responsible for omitting such attribute-values from the list of attributes
        # in a NodeProto.
        if val is None:
            if attr_meta and attr_meta.required:
                self._fail(expr, f"Attribute '{attr_name}' is required.")
            return None
        attr_type = attr_meta.type if attr_meta else None
        if attr_type == ir.AttributeType.TENSOR:
            val = ir.tensor(val)
        attr = ir.convenience.convert_attribute(attr_name, val, attr_type)
        return attr

    def _translate_docstring(self, node: ast.Expr) -> None:
        if not isinstance(node.value, ast.Constant):
            self._fail(node, "Docstring expression must be a constant.")
        self._current_fn.doc_string = node.value.value

    def _translate_expr(self, node: ast.AST, target: PreferredName | None = None) -> ir.Value:
        """Expression-translation generates "IR statements/nodes" that compute the value of
        the expression into a target-variable, and returns the variable that is
        assigned this value.
        """
        if isinstance(node, ast.Call):
            r = self._translate_call_expr(node)
        elif isinstance(node, (ast.BinOp, ast.BitAnd, ast.BitOr)):
            r = self._translate_binary_op_expr(node)
        elif isinstance(node, ast.UnaryOp):
            r = self._translate_unary_op_expr(node)
        elif isinstance(node, ast.Compare):
            r = self._translate_compare_expr(node)
        elif isinstance(node, ast.Name):
            r = self._translate_name_expr(node)
        elif isinstance(node, ast.Subscript):
            r = self._translate_subscript_expr(node, target)
        elif self._is_constant_expr(node):
            r = self._emit_const(self._eval_constant_expr(node), target, self._source_of(node))
        else:
            raise ValueError(
                self._message(node, f"Unsupported expression type {type(node)!r}.")
            )
        if isinstance(r, ir.Value):
            return r
        callee, args, attrs = r
        target = "tmp" if target is None else target
        assert isinstance(target, str)
        result = self._generate_unique_name(target)
        return self._emit1([result], callee, args, attrs)

    def _translate_opt_expr(self, node: ast.expr) -> ir.Value | None:
        """Translation of an expression where "None" is permitted (eg., for an optional argument).
        None is represented as a Constant in Python 3.9+.
        """
        if isinstance(node, ast.Constant) and (node.value is None):
            return None
        return self._translate_expr(node)

    def _translate_subscript_expr(
        self, node: ast.Subscript, target: PreferredName | None
    ) -> ir.Value:
        """List of supported syntaxes is below.
        `A` is a tensor or an expression equivalent to a tensor.

        ::

            A[:, 1]
            A[:2, 0]
            A[:2, :1]
            A[2:0:-1]
            A[1:]
            A[:2]
            A[1:-1]
            A[1:2]
            A[-1]
            A[0]
            A[:0:-1]

        *i* is a tensor holding one integer.

        ::

            A[i]
            A[i+1:i+2]

        Fully supported for python 3.9+.

        ::

            A[i:i+j, k]

        Not supported:

        ::

            A[::-1]
        """
        var = self._translate_expr(node.value)
        var_name = var.name
        if target is None:
            target = f"{var_name}_subscripted"
        target = self._generate_unique_name(target)
        indices = ast_utils.normalize_subscript_expr(node)
        info = self._source_of(node.slice)

        # Create cached int constants:
        # TODO: Do this at a graph-scope level.
        cached_int_consts: dict[int, ir.Value] = {}

        def const_1d(value, name: str | None = None) -> ir.Value:
            nonlocal cached_int_consts
            if value not in cached_int_consts:
                cached_int_consts[value] = self._emit_const([value], name, info)
            return cached_int_consts[value]

        def one_1d() -> ir.Value:
            return const_1d(1)

        # Max/min 64-bit int values are used to represent default values for start/stop in Slice.
        maxint = (1 << 63) - 1
        minint = -(1 << 63)

        def translate_slice_component(
            node_arg, default_value: int | None = None
        ) -> tuple[ir.Value, int | None]:
            """Translate optional start/stop/step component of a Slice expression."""
            if node_arg is None:
                if default_value is None:
                    # TODO: Emit "Where(step > 0, pos_default, neg_default)"
                    raise RuntimeError(
                        "Default start/stop not supported when step direction is unknown."
                    )
                return const_1d(default_value), default_value

            if self._is_constant_expr(node_arg):
                cst = self._eval_constant_expr(node_arg)
                if isinstance(cst, int):
                    return const_1d(cst), cst
                else:
                    raise RuntimeError(f"Slice component type must be int, not {type(cst)}")
            else:
                value = self._translate_expr(node_arg)
                reshaped = self._generate_unique_name(f"{value.name}_reshaped")
                reshaped_value = self._emit1(
                    [reshaped],
                    values.Op(self.default_opset, "Reshape"),
                    [value, one_1d()],
                    [],
                )
                return reshaped_value, None

        def translate_slice(slice_expr: ast.Slice) -> tuple[ir.Value, ir.Value, ir.Value]:
            """Translate slice-expression of the form from:to:step."""
            step_name, step = translate_slice_component(slice_expr.step, 1)
            if step is None:
                # Step direction unknown.
                # TODO: Handle default-values using runtime check on sign of step.
                lower_name, _ = translate_slice_component(slice_expr.lower, None)
                upper_name, _ = translate_slice_component(slice_expr.upper, None)
            elif step > 0:
                lower_name, _ = translate_slice_component(slice_expr.lower, 0)
                upper_name, _ = translate_slice_component(slice_expr.upper, maxint)
            else:
                lower_name, _ = translate_slice_component(slice_expr.lower, maxint)
                upper_name, _ = translate_slice_component(slice_expr.upper, minint)
            return (lower_name, upper_name, step_name)

        # An input like X[2] is translated into a Gather op.
        # An input like X[1:5:2] is translated into a Slice op.
        # An input like X[2, 3] is translated into a Slice + Squeeze (instead of two Gathers),
        #   as an optimization.
        # An input like X[I, J] is translated into two Gathers (which is correct whatever the
        #   rank of I and J)
        # To replace multiple Gathers by the Slice we need to know that the index-values
        # are scalars.

        # As the first step, we partition the index elements into four kinds: Slice (eg., 1:5:2),
        # known-to-be-scalar (eg., 2), other-tensor (eg., I), skip/no-op (that is, just ":")
        sliced_indices: list[tuple[int, ast.expr]] = []
        scalar_indices: list[tuple[int, ast.expr]] = []
        non_scalar_indices: list[tuple[int, ast.expr]] = []
        for axis, elt in enumerate(indices):
            if isinstance(elt, ast.Slice):
                # Add to sliced_indices, unless it is "::", which is a no-op.
                if not (elt.lower is None and elt.upper is None and elt.step is None):
                    sliced_indices.append((axis, elt))
            elif self._is_constant_expr(elt) and isinstance(
                self._eval_constant_expr(elt), int
            ):
                scalar_indices.append((axis, elt))
            else:
                non_scalar_indices.append((axis, elt))
        if not (sliced_indices or scalar_indices or non_scalar_indices):
            # Edge case: no index specified. Eg. A[:, :]
            return self._emit1([target], "Identity", [var_name])

        if sliced_indices or len(scalar_indices) > 1:
            # We emit a Slice operation if we have any indices like 1:5:2 or if the number of
            # scalar indices (like 2) is more than 1.
            starts = []
            ends = []
            axes = []
            steps = []
            squeezed_axes = []
            for axis, expr in scalar_indices:
                # Treat a scalar index i as slice "i:i+1:1", but squeeze the axis finally.
                # TODO: handle negative i
                index = self._eval_constant_expr(expr)
                squeezed_axes.append(axis)
                kwargs = dict(
                    lineno=getattr(expr, "lineno", node.lineno),
                    col_offset=getattr(expr, "col_offset", node.col_offset),
                )
                element = ast.Slice(
                    ast.Constant(index, **kwargs),
                    ast.Constant(index + 1, **kwargs),
                    ast.Constant(1, **kwargs),
                )
                sliced_indices.append((axis, element))
            scalar_indices = []
            for axis, element in sliced_indices:
                axis_var = const_1d(axis)
                inputs = translate_slice(element)
                starts.append(inputs[0])
                ends.append(inputs[1])
                axes.append(axis_var)
                steps.append(inputs[2])

            if len(starts) > 1:
                axis_0_attr = ir.AttrInt64("axis", 0)
                start_name = self._generate_unique_name(f"{var_name}_start")
                start_value = self._emit([start_name], "Concat", starts, [axis_0_attr])

                end_name = self._generate_unique_name(f"{var_name}_end")
                end_value = self._emit([end_name], "Concat", ends, [axis_0_attr])

                axes_name = self._generate_unique_name(f"{var_name}_axis")
                axes_value = self._emit([axes_name], "Concat", axes, [axis_0_attr])

                steps_name = self._generate_unique_name(f"{var_name}_step")
                steps_value = self._emit([steps_name], "Concat", steps, [axis_0_attr])
            else:
                start_value = starts[0]
                end_value = ends[0]
                axes_value = axes[0]
                steps_value = steps[0]

            if squeezed_axes:
                sliced_name = self._generate_unique_name(f"{var_name}_sliced")
                sliced_value = self._emit(
                    [sliced_name],
                    "Slice",
                    [var, start_value, end_value, axes_value, steps_value],
                )
                squeezed_axes = self._emit_const(squeezed_axes, "squeezed_axes", info)

                if non_scalar_indices:  # use temporary to store result of squeeze
                    result_name = self._generate_unique_name(f"{var_name}_squeezed")
                else:  # store squeezed result in final target
                    result_name = target

                result = self._emit([result_name], "Squeeze", [sliced_value, squeezed_axes])
            else:
                if non_scalar_indices:  # use temporary to store result of Slice
                    result_name = self._generate_unique_name(f"{var_name}_sliced")
                else:  # store result of Slice in final target
                    result_name = target
                slice_inputs = [var, start_value, end_value, axes_value, steps_value]
                result = self._emit1([result_name], "Slice", slice_inputs)
        else:
            result = var
        non_scalar_indices.extend(scalar_indices)
        if non_scalar_indices:
            last_axis, _ = non_scalar_indices[-1]
        else:
            # TODO(justinchuby): Clarify what last_axis should be when non_scalar_indices is False
            last_axis = None
        for axis, index_expr in non_scalar_indices:
            index_value = self._translate_expr(index_expr)
            axis_attr = ir.AttrInt64("axis", axis)
            # use Gather to perform indexing
            # Assign gathered value to either temporary or final target
            if axis != last_axis:  # use temporary to store result of Gather
                gathered = self._generate_unique_name(f"{var_name}_axis_{axis}")
            else:  # store result of Gather in final target
                gathered = target
            result = self._emit1([gathered], "Gather", [result, index_value], [axis_attr])

        return result

    def _translate_call_expr(
        self, node: ast.Call
    ) -> tuple[values.Op, list[ir.Value | None], list[ir.Attr]]:
        """Translates a call-expression."""
        callee = self._translate_callee_expr(node.func)
        op_signature = callee.op_signature
        # If the callee's schema is available, we use it to determine the inputs and attributes.
        # Otherwise, we map named arguments to attributes and positional arguments to inputs.
        if op_signature:
            kwargs = {x.arg: x.value for x in node.keywords}
            args, attrs = param_manipulation.separate_input_attributes_from_arguments(
                op_signature, node.args, kwargs, fill_defaults=False
            )
            args = [self._translate_opt_expr(x) for x in args]
            attrs = [self._translate_attr(x, y, op_signature.get(x)) for x, y in attrs.items()]
        else:
            args = [self._translate_opt_expr(x) for x in node.args]
            attrs = [self._translate_attr(x.arg, x.value) for x in node.keywords]
        args = autocast.static_cast_inputs(self, op_signature, args)

        # In ONNX, there is no way to explicitly specify a None value for an attribute.
        # Instead, the attribute must be omitted from the attribute list.
        # Hence, we do not create an attribute-proto if the value is None.
        attrs = [attr for attr in attrs if attr is not None]
        return callee, args, attrs

    def _cast_like_binary_expression(self, op, left, right) -> tuple[ir.Value, ir.Value]:
        op_signature = op.op_signature
        return autocast.static_cast_inputs(self, op_signature, (left, right))

    def _translate_binary_op_expr(self, node: ast.BinOp):
        op = type(node.op)
        if op not in primop_map:
            raise ValueError(self._message(node, f"Unsupported operator {op!r}."))

        attrs = []
        if isinstance(node.op, ast.Mod) and self._is_constant_expr(node.right):
            # specific case X % f where f is a float.
            # attribute fmod=1 is added in that case.
            cst = self._eval_constant_expr(node.right)
            if isinstance(cst, float):
                attrs = [ir.AttrInt64("fmod", 1)]

        op = values.Op(self.default_opset, primop_map[op])
        left, right = self._cast_like_binary_expression(
            op, self._translate_expr(node.left), self._translate_expr(node.right)
        )
        return op, [left, right], attrs

    def _translate_unary_op_expr(self, node):
        op = type(node.op)
        if op not in primop_map:
            raise ValueError(self._message(node, self).msg(f"Unsupported operator {op!r}."))
        if self._is_constant_expr(node.operand):
            # This function changed the constant node.operand
            # and returns it. The function calling this one
            # should intercept this call and replace node
            # by node.operand.
            # This mechanism does not handle somthing like `(-(-5))`.
            val = node.operand.value
            if op == ast.USub:
                cst = ast.Constant(-val, lineno=node.lineno, col_offset=node.col_offset)
                return self._translate_expr(cst)
            if op == ast.UAdd:
                return self._translate_expr(node.operand)
        opname = primop_map[op]
        operand = self._translate_expr(node.operand)
        return values.Op(self.default_opset, opname), [operand], []

    def _translate_compare_expr(self, node):
        # TODO: handle multiple comparisons in one expression
        if len(node.ops) != 1 or len(node.comparators) != 1:
            self._fail(node, "Multiple comparisons in one expression are not supported.")
        op = type(node.ops[0])
        if op not in primop_map:
            raise ValueError(self._message(node, f"Unsupported operator {op!r}."))
        opname = primop_map[op]
        left = self._translate_expr(node.left)
        right = self._translate_expr(node.comparators[0])

        # NotEqual is not a standard ONNX op, and needs to be translated into
        # an Equal op/node followed by a Not op/node.
        op = values.Op(self.default_opset, opname if opname != "NotEqual" else "Equal")
        left, right = self._cast_like_binary_expression(op, left, right)
        if opname == "NotEqual":
            tmp = self._generate_unique_name()
            tmp_value = self._emit1([tmp], op, [left, right])
            not_op = values.Op(self.default_opset, "Not")
            return not_op, [tmp_value], []

        return op, [left, right], []

    def _translate_name_expr(self, node: ast.Name) -> ir.Value:
        return self._py_var_to_onnx_var(node.id, self._source_of(node))

    def _translate_opset_expr(self, node: ast.Attribute) -> values.Opset:
        """Return an Opset"""
        if isinstance(node, ast.Name):
            val = self._lookup(node.id, self._source_of(node), raise_exception=False)
            if isinstance(val, values.Opset):
                return val
            elif isinstance(val, builder.OpBuilder):
                # Convert OpBuilder to Opset for compatibility
                return values.Opset(val.domain, val.version)
            self._fail(node, f"'{node.id}' is not an instance of type Opset but {type(val)}.")
        elif isinstance(node, ast.Attribute):
            self._fail(node, "Nested module unimplemented.")  # TODO
        else:
            self._fail(node, "Invalid opset expression.")

    def _translate_callee_expr(self, node: ast.AST) -> values.Op:  # pylint: disable=R1710
        """Return an Op"""
        if isinstance(node, ast.Attribute):
            module = self._translate_opset_expr(node.value)
            self._set_default_opset(module, node)
            return values.Op(module, node.attr)
        if isinstance(node, ast.Name):
            function_name = node.id
            found = self._lookup(function_name, self._source_of(node), raise_exception=False)
            if isinstance(found, (values.Op, onnxscript.OnnxFunction)):
                return found
            if not found:
                if function_name not in self.default_opset:
                    warn(
                        f"Unknown function name {function_name!r}. "
                        f"The ONNX graph may not work."
                    )
                return values.Op(self.default_opset, function_name)
        self._fail(node, "Invalid callee")

    def _translate_stmt(self, node: ast.stmt, index_of_stmt=None) -> None:
        """Statement translation: A single Python statement is mapped into a
        sequence of IR statements.
        """
        if isinstance(node, ast.Assign):
            return self._translate_assign_stmt(node)
        if isinstance(node, ast.AnnAssign):
            return self._translate_assign_stmt(node)
        if isinstance(node, ast.Return):
            if index_of_stmt is not None:
                return self._translate_return_stmt(node)
            raise ValueError(
                self._message(
                    node, "Return statements are not permitted inside control-flow statements."
                )
            )
        if isinstance(node, ast.If):
            return self._translate_if_stmt(node)
        if isinstance(node, (ast.For, ast.While)):
            return self._translate_loop_stmt(node)
        if ast_utils.is_doc_string(node):
            if index_of_stmt == 0:
                return self._translate_docstring(node)
            return None
        if isinstance(node, ast.FunctionDef):
            return self._translate_nested_function_def(node)
        if ast_utils.is_print_call(node):
            return None
        raise ValueError(self._message(node, f"Unsupported statement type '{type(node)!r}'."))

    def _translate_assign_stmt(self, stmt: ast.Assign | ast.AnnAssign) -> None:
        def assign(lhs: ast.AST, rhs: ast.AST) -> None:
            if isinstance(lhs, ast.Name):
                # Assignments of the form "x = SomeExpression"
                info = self._source_of(lhs)
                lhs = lhs.id
                t = self._translate_expr(rhs, lhs)
                if isinstance(stmt, ast.AnnAssign):
                    typeinfo = self._get_type_annotation(stmt.annotation)
                else:
                    typeinfo = None
                if typeinfo is not None:
                    set_type_info(t, typeinfo)
                var = values.SymbolValue(t, info)
                self._bind(lhs, var)
            elif isinstance(lhs, ast.Tuple):
                # Assignments of the form "x, y, z = op.SomeOp(...)"
                if not isinstance(rhs, ast.Call):
                    self._fail(
                        rhs,
                        f"RHS must be a Call expression for unpacking, found: '{type(rhs)!r}'",
                    )
                callee, inputs, attrs = self._translate_call_expr(rhs)

                def generate_onnx_name(x: ast.AST):
                    if not isinstance(x, ast.Name):
                        self._fail(
                            x, f"LHS must be a Name for unpacking, found: '{type(x)!r}'"
                        )
                    return self._generate_unique_name(x.id)

                output_names = [generate_onnx_name(x) for x in lhs.elts]
                outputs = self._emit(output_names, callee, inputs, attrs)
                if isinstance(outputs, ir.Value):
                    outputs = [outputs]
                for x, output in zip(lhs.elts, outputs):
                    self._bind(
                        x.id,
                        values.SymbolValue(output, self._source_of(x)),
                    )
            else:
                self._fail(lhs, f"Unsupported construct in LHS of assignment: '{type(lhs)!r}'")

        if isinstance(stmt, ast.Assign):
            targets = stmt.targets
        else:
            targets = [stmt.target]
        if len(targets) != 1:
            # Assignments of the form "x = y = SomeExpression"
            self._fail(stmt, "Multi-assignment not supported.")
        lhs = targets[0]
        rhs = stmt.value
        if isinstance(rhs, ast.Tuple):
            # Assignments of the form "... = Expression1, Expression2"
            if not isinstance(lhs, ast.Tuple):
                # Assignments of the form "single_var = Expression1, Expression2".
                # We do not support tuple-typed variables.
                self._fail(lhs, f"Left term must be a tuple not '{type(lhs)!r}'.")
            # Parallel assignments of the form "x, y = Expression1, Expression2"
            if len(lhs.elts) != len(rhs.elts):
                self._fail(
                    stmt, "Expected same number of elements on lhs and rhs of assignments."
                )
            for p, r in zip(lhs.elts, rhs.elts):
                assign(p, r)
        else:
            assign(lhs, rhs)

    def _translate_return_stmt(self, stmt: ast.Return) -> None:
        def check_num_outputs(n):
            if self.returntype is not None:
                if n != len(self.returntype):
                    raise SyntaxError(
                        self._message(
                            stmt,
                            f"Mismatch in number of return values and types. Keyword "
                            f"'return' cannot be used in a subgraph (test, loop).  "
                            f"returntype is {self.returntype!r}, num_outputs={n!r}.",
                        )
                    )

        def ret(exp, i, suffix):
            preferred_name = f"return_val{suffix}"
            return_var = self._translate_expr(exp, preferred_name)
            val = self._lookup(return_var.name, self._source_of(exp), raise_exception=False)
            if isinstance(val, values.SymbolValue) and isinstance(val.value, ir.Value):
                if val.value.is_graph_input():
                    # In ONNX, a graph-input cannot be an output of the graph.
                    # We need to insert a copy.
                    return_var = self._emit_copy(return_var, preferred_name)
            for prev_output in self._current_fn.outputs:
                if prev_output.name == return_var.name:
                    # ONNX does not allow duplicate output names.
                    return_var = self._emit_copy(return_var, f"{return_var}_copy")
                    break
            if self.returntype is not None:
                set_type_info(return_var, self.returntype[i])
            self._current_fn.outputs.append(return_var)
            return return_var

        val = stmt.value
        if val is None:
            self._fail(stmt, "Return statement without a return value is not supported.")
        if isinstance(val, ast.Tuple):
            check_num_outputs(len(val.elts))
            return [ret(exp, i, str(i)) for i, exp in enumerate(val.elts)]
        check_num_outputs(1)
        return ret(val, 0, "")

    def _translate_if_stmt(self, stmt: ast.If) -> None:
        constant_cond = self.analyzer.constant_if_condition(stmt)
        if constant_cond is True:
            # Translate only the "then" branch
            for s in stmt.body:
                self._translate_stmt(s)
            return
        if constant_cond is False:
            # Translate only the "else" branch
            for s in stmt.orelse:
                self._translate_stmt(s)
            return
        live_def_set = self.analyzer.assigned_vars(stmt)
        live_out = self.analyzer.live_out(stmt)
        if live_out is not None:
            # Ideally, live_out should never be None here. But handle this conditionally
            # due to some existing usage.
            live_def_set = live_out.intersection(live_def_set)
        live_defs = list(live_def_set)
        test = self._translate_expr(stmt.test, "cond")
        lineno = self._source_of(stmt).lineno
        then_graph = self._translate_block(stmt.body, f"thenGraph_{lineno}", live_defs)
        then_attr = ir.AttrGraph("then_branch", then_graph)
        else_graph = self._translate_block(stmt.orelse, f"elseGraph_{lineno}", live_defs)
        else_attr = ir.AttrGraph("else_branch", else_graph)

        def rename(x):
            return self._generate_unique_name(x)

        # no break condition
        renamed = [rename(x) for x in live_defs]
        if not renamed:
            self._fail(stmt, "A subgraph for a test do not have any output variable.")

        if renamed == [test.name]:
            self._fail(stmt, f"Input and output cannot be the same {renamed!r}.")
        if_outputs = self._emit(
            renamed,
            values.Op(self.default_opset, "If"),
            [test],
            [then_attr, else_attr],
        )
        if isinstance(if_outputs, ir.Value):
            if_outputs = [if_outputs]
        for x, y in zip(live_defs, if_outputs):
            self._bind(
                x,
                values.SymbolValue(y, self._source_of(stmt)),
            )

    def _translate_loop_stmt(self, loop_stmt: ast.For | ast.While) -> None:
        # loop-variable
        if isinstance(loop_stmt, ast.For):
            if not isinstance(loop_stmt.target, ast.Name):
                self._fail(loop_stmt, "For loop target must be a single variable.")
            python_loop_var_name = loop_stmt.target.id
            # iter
            iter = loop_stmt.iter
            if not isinstance(iter, ast.Call):
                self._fail(loop_stmt, "Loop bound must be a function call (e.g. 'range(n)').")
            if not isinstance(iter.func, ast.Name):
                self._fail(loop_stmt, f"Unsupported loop bound {iter.func!r}.")
            if iter.func.id != "range":
                self._fail(
                    loop_stmt, "Unsupported loop bound, only function 'range' is allowed."
                )
            if not iter.args or len(iter.args) != 1:
                self._fail(loop_stmt, "Unsupported loop bound, it should be 'range(?)'.")
            if iter.keywords:
                self._fail(loop_stmt, "Keyword arguments in loop bound are not supported.")
            o_loop_bound = self._translate_expr(iter.args[0], "loop_bound")
            onnx_cond_var = make_value(
                self._generate_unique_name("cond_in"),
                onnx_types.BOOL,
                self._source_of(loop_stmt),
            )
            i_cond_var = onnx_cond_var
            cond_while = None
            o_loop_condition = None  # No condition for a for loop.
        elif isinstance(loop_stmt, ast.While):
            test = loop_stmt.test
            if not isinstance(test, ast.Name):
                self._fail(
                    loop_stmt,
                    "Unexpected condition type {type(loop_stmt)!r} for a while loop, "
                    "it should be 'while <condition_name>:'.",
                )
            python_loop_var_name = "infinite_loop"
            o_loop_bound = None
            i_cond_var = make_value(
                self._generate_unique_name(test.id),
                onnx_types.BOOL,
                self._source_of(loop_stmt),
            )
            cond_while = test.id
            onnx_cond_var = None
            o_loop_condition = self._translate_name_expr(test)
            # we need to go through all the instructions to see
            # which instruction defines the condition test.id
        else:
            self._fail(loop_stmt, f"Unexpected loop type {type(loop_stmt)!r}.")
        # analyze loop body
        exposed_uses = self.analyzer.exposed_uses(loop_stmt.body)
        vars_def_in_loop = self.analyzer.assigned_vars(loop_stmt.body)
        live_out = self.analyzer.live_out(loop_stmt)
        assert live_out is not None, "live_out cannot be None here."
        loop_state_vars = vars_def_in_loop.intersection(exposed_uses | live_out)
        scan_outputs = set()  # TODO
        outputs = list(loop_state_vars | scan_outputs)

        # loop-condition:
        # o_loop_condition = self._emit_const(True, "true", self._source_of(loop_stmt))

        # build loop_body
        self._enter_scope("loop_body", loop_stmt)
        onnx_loop_var_name = self._generate_unique_name(python_loop_var_name)
        onnx_loop_var = make_value(
            onnx_loop_var_name,
            onnx_types.INT64,
            self._source_of(loop_stmt),
        )
        self._current_fn.append_parameter(onnx_loop_var)
        self._bind(
            python_loop_var_name,
            values.SymbolValue(onnx_loop_var, self._source_of(loop_stmt)),
        )

        self._current_fn.append_parameter(i_cond_var)

        for pv in loop_state_vars:
            onnx_var_name = self._generate_unique_name(pv)
            parameter = make_value(onnx_var_name, None, self._source_of(loop_stmt))
            self._current_fn.append_parameter(parameter)
            self._bind(
                pv,
                values.SymbolValue(
                    parameter,
                    self._source_of(loop_stmt),
                ),
            )

        condition_name: ir.Value | None = None
        operator_name = "Identity"
        for i, s in enumerate(loop_stmt.body):
            # We first need to intercept a break instruction in test block.
            # It must be something like `if <condition_name>: break`.
            # This instruction must be the last of the loop body.
            if isinstance(s, ast.If) and len(s.body) == 1 and isinstance(s.body[0], ast.Break):
                if not isinstance(s.test, ast.Name):
                    self._fail(
                        s,
                        f"Instruction break can be introduced with test but it must be "
                        f"if <condition>: break. However condition is of type "
                        f"{type(s.test)!r}.",
                    )
                if i != len(loop_stmt.body) - 1:
                    self._fail(s, "Instruction break must be the last one of the loop.")

                current_scope = self._current_scope()
                if s.test.id not in current_scope:
                    self._fail(
                        loop_stmt,
                        f"Unable to find condition variable {s.test.id!r} in known "
                        f"variables {list(current_scope)!r}.",
                    )
                condition_name = current_scope[s.test.id].value
                operator_name = "Not"
                continue
            self._translate_stmt(s)

        if cond_while is not None:
            # Loop while
            current_scope = self._current_scope()
            if cond_while not in current_scope:
                self._fail(
                    loop_stmt,
                    f"Unable to find condition variable {cond_while} in known "
                    f"variables {list(current_scope)!r}.",
                )
            onnx_cond_var = current_scope[cond_while].value

        cond_out = self._emit1(
            [self._generate_unique_name("cond_out")],
            values.Op(self.default_opset, operator_name),
            [condition_name or onnx_cond_var],
            [],
        )
        self._current_fn.outputs.append(cond_out)

        for pv in loop_state_vars:
            onnx_var = self._py_var_to_onnx_var(pv, self._source_of(loop_stmt))
            if onnx_var.name not in self._current_fn.assigned_names:
                # When converting the loop-body into a graph, we need to handle
                # identity assignments of the form "x = y" inside the loop body
                # specially if y represents a value computed outside the loop body.
                # In this case, we create a copy of y, treating the statement as
                # shorthand for "x = op.Identity(y)".
                onnx_var = self._emit_copy(onnx_var, pv)
            self._current_fn.outputs.append(onnx_var)
        body = self._exit_scope()
        inputs = [o_loop_bound, o_loop_condition] + [
            self._py_var_to_onnx_var(pv, self._source_of(loop_stmt)) for pv in loop_state_vars
        ]
        attrs = [ir.AttrGraph("body", body.graph)]
        info = self._source_of(loop_stmt)

        def rename(x):
            r = self._generate_unique_name(x)
            return r

        onnx_output_names = [rename(x) for x in outputs]
        loop_outputs = self._emit(
            onnx_output_names,
            "Loop",
            inputs,
            attrs,
        )
        if isinstance(loop_outputs, ir.Value):
            loop_outputs = [loop_outputs]
        for x, loop_output in zip(outputs, loop_outputs):
            self._bind(x, values.SymbolValue(loop_output, info))

    def _translate_block(
        self,
        stmts: Sequence[ast.stmt],
        name: str,
        live_defs: Sequence[str],
    ) -> ir.Graph:
        """Translation of a then/else statement-block to an ir.Graph."""
        self._enter_scope(name, None)
        for s in stmts:
            self._translate_stmt(s)
        for python_var in live_defs:
            if python_var in self._current_scope():
                python_var_value = self._current_scope()[python_var]
                output = self._to_onnx_var(python_var_value, python_var)
                if output.name not in self._current_fn.assigned_names:
                    # TODO (Rama): Unclear how this can happen. If python_var is in current_scope,
                    # then it should have been assigned a value in the current graph.
                    #
                    # To return an outer-scope variable, an ONNX Graph has to
                    # use an explicit copy via Identity.
                    output = self._emit_copy(output, python_var)
                self._current_fn.outputs.append(output)
            else:
                python_var_value = None
                for scope in reversed(self._locals):  # TODO: skip _current_scope
                    if python_var in scope:
                        python_var_value = scope[python_var]
                        break
                if python_var_value is None:
                    self._fail(
                        stmts[0],
                        f"ir.Value {python_var} is not assigned a value along a conditional "
                        f"branch, known variables: {list(self._locals)}.",
                    )
                # introduce a copy
                output = self._emit_copy(
                    self._to_onnx_var(python_var_value, python_var), python_var
                )

                # TODO: retrieve the annotation if any.
                self._current_fn.outputs.append(output)
        function_ir = self._exit_scope()
        return function_ir.graph

    def _translate_nested_function_def(self, fn: ast.FunctionDef) -> None:
        """Translate a nested function definition."""
        self._enter_scope(fn.name, fn)
        self._translate_function_def_common(fn)
        function_ir = self._exit_scope()
        outer_scope_vars = self.analyzer.outer_scope_variables(fn)
        function_ir.outer_scope_variables = [
            (var, self._lookup(var, self._source_of(fn))) for var in outer_scope_vars
        ]
        self._bind(fn.name, values.SymbolValue(function_ir, self._source_of(fn)))
        # TODO: Does not yet handle nested functions within nested functions.
        self._current_fn.add_nested_function(function_ir)

    def _translate_function_signature_common(
        self, fn: ast.FunctionDef
    ) -> irbuilder.IRFunction:
        """Translate a function signature (top-level or nested)."""
        args = fn.args
        if args.vararg or args.kwonlyargs or args.kw_defaults or args.kwarg:
            warn(f"{fn.name}: Unsupported feature in function signature.")
        for i, x in enumerate(args.args):
            arg_with_default_start_index = len(args.args) - len(args.defaults)
            if args.defaults and i >= arg_with_default_start_index:
                default_value = self._eval_constant_expr(
                    args.defaults[i - arg_with_default_start_index]
                )
            else:
                default_value = None
            if x.annotation:
                typeinfo = self._get_type_annotation(x.annotation)
            else:
                # The code can only be exported as a function.
                typeinfo = None
            if typeinfo and ta.is_attr_type(typeinfo):
                attribute_type = _schemas.get_attr_type(typeinfo)
                attr = ir.Attr(x.arg, attribute_type, default_value, None)
                self._current_fn.append_parameter(attr)
                as_bool = ta.base_type_is_bool(typeinfo)
                self._bind(x.arg, values.AttrRef(attr, as_bool, self._source_of(x)))
            else:
                onnx_parameter = make_value(x.arg, typeinfo, self._source_of(x))
                self._current_fn.append_parameter(onnx_parameter)
                self._used_vars.add(x.arg)
                self._bind(
                    x.arg,
                    values.SymbolValue(onnx_parameter, self._source_of(x)),
                )
        if fn.returns:
            type_annotation = self._eval_constant_expr(fn.returns)
            self.returntype = ta.get_return_types(type_annotation)
            invalid = False
            for t in self.returntype:
                if not ta.is_valid_type(t):
                    self._warn(
                        fn.returns,
                        f"Unsupported type annotation for return value {t}.",
                    )
                    invalid = True
            if invalid:
                self.returntype = None
        else:
            self.returntype = None

        return self._current_fn

    def _translate_function_def_common(self, fn: ast.FunctionDef) -> irbuilder.IRFunction:
        """Translate a function definition, including the signature and its body."""
        logger.debug("Converter:_translate_function_def_common:%s", fn.name)
        _ = self._translate_function_signature_common(fn)
        for i, s in enumerate(fn.body):
            self._translate_stmt(s, index_of_stmt=i)
        return self._current_fn

    def translate_function_def(self, stmt: ast.FunctionDef) -> irbuilder.IRFunction:
        if isinstance(stmt, ast.FunctionDef):
            self._init_function_translation()
            if self.default_opset_ is None:
                opset = self._find_onnx_opset(stmt)
                if opset:
                    self._set_default_opset(opset, stmt)
            domain = self.this_module.domain
            self._current_fn = irbuilder.IRFunction(stmt.name, domain)
            self._analyzer = analysis.AstAnalyzer(stmt, self._message, self.globals)
            fn_ir = self._translate_function_def_common(stmt)
            self.this_module.add_function_def(fn_ir)
            self._analyzer = None
            return fn_ir
        raise ValueError(f"Unsupported top-level statement type {type(stmt)!r}.")

    def translate_function_signature(self, fn: ast.FunctionDef) -> irbuilder.IRFunction:
        """Translate a (top-level) function signature."""
        domain = self.this_module.domain
        self._current_fn = irbuilder.IRFunction(fn.name, domain)
        return self._translate_function_signature_common(fn)
